# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

import math
import os

import kaldi_native_fbank as knf
import kaldiio
import numpy as np
import soundfile as sf
import torch


class AudioFeat:
    def __init__(self, kaldi_cmvn_file):
        self.cmvn = CMVN(kaldi_cmvn_file) if kaldi_cmvn_file != "" else None
        self.fbank = KaldifeatFbank(
            num_mel_bins=80, frame_length=25, frame_shift=10, dither=0
        )

    def reset(self):
        pass

    def extract(self, audio):
        if isinstance(audio, str):
            wav_np, sample_rate = sf.read(audio, dtype="int16")
        elif isinstance(audio, (list, tuple)):
            wav_np, sample_rate = audio
        else:
            wav_np = audio
            sample_rate = 16000
        assert sample_rate == 16000

        dur = wav_np.shape[0] / sample_rate
        fbank = self.fbank((sample_rate, wav_np))
        if self.cmvn is not None:
            fbank = self.cmvn(fbank)
        feat = torch.from_numpy(fbank).float()
        return feat, dur


class CMVN:
    def __init__(self, kaldi_cmvn_file):
        self.dim, self.means, self.inverse_std_variances = self.read_kaldi_cmvn(
            kaldi_cmvn_file
        )

    def __call__(self, x, is_train=False):
        assert x.shape[-1] == self.dim, "CMVN dim mismatch"
        out = x - self.means
        out = out * self.inverse_std_variances
        return out

    def read_kaldi_cmvn(self, kaldi_cmvn_file):
        assert os.path.exists(kaldi_cmvn_file)
        stats = kaldiio.load_mat(kaldi_cmvn_file)
        assert stats.shape[0] == 2
        dim = stats.shape[-1] - 1
        count = stats[0, dim]
        assert count >= 1
        floor = 1e-20
        means = []
        inverse_std_variances = []
        for d in range(dim):
            mean = stats[0, d] / count
            means.append(mean.item())
            variance = (stats[1, d] / count) - mean * mean
            if variance < floor:
                variance = floor
            istd = 1.0 / math.sqrt(variance)
            inverse_std_variances.append(istd)
        return dim, np.array(means), np.array(inverse_std_variances)


class KaldifeatFbank:
    def __init__(self, num_mel_bins=80, frame_length=25, frame_shift=10, dither=0):
        self.dither = dither
        opts = knf.FbankOptions()
        opts.frame_opts.samp_freq = 16000
        opts.frame_opts.frame_length_ms = 25
        opts.frame_opts.frame_shift_ms = 10
        opts.frame_opts.dither = dither
        opts.frame_opts.snip_edges = True
        opts.mel_opts.num_bins = num_mel_bins
        opts.mel_opts.debug_mel = False
        self.opts = opts

    def __call__(self, wav, is_train=False):
        if isinstance(wav, str):
            wav_np, sample_rate = sf.read(wav, dtype="int16")
        elif isinstance(wav, (tuple, list)) and len(wav) == 2:
            sample_rate, wav_np = wav
        assert len(wav_np.shape) == 1

        dither = self.dither if is_train else 0.0
        self.opts.frame_opts.dither = dither
        fbank = knf.OnlineFbank(self.opts)

        fbank.accept_waveform(sample_rate, wav_np.tolist())
        feat = []
        for i in range(fbank.num_frames_ready):
            feat.append(fbank.get_frame(i))
        if len(feat) == 0:
            return np.zeros((0, self.opts.mel_opts.num_bins))
        feat = np.vstack(feat)
        return feat



from fasr_asr_paraformer.models.asr import ParaformerForASR, SeacoParaformerForASR
from fasr_asr_paraformer.models.stream_asr import (
    ParaformerForStreamASR,
    ParaformerForStreamASROnnx,
)

__all__ = [
    "ParaformerForASR",
    "SeacoParaformerForASR",
    "ParaformerForStreamASR",
    "ParaformerForStreamASROnnx",
]



from pathlib import Path
from typing import List, Optional

from loguru import logger
import numpy as np
import onnxruntime as ort

from fasr.config import Config, registry

from .base import BaseRuntime


class ORT(BaseRuntime):
    onnx_model_path: str | Path | None = None
    device_id: int | None = None
    intra_op_num_threads: int = 2
    session: ort.InferenceSession | None = None
    verbose: bool = True

    def __call__(self, input_content: List[np.ndarray]) -> np.ndarray:
        return self.run(input_content)

    def run(self, input_content: List[np.ndarray]) -> np.ndarray:
        if self.session is None:
            raise ONNXRuntimeError("ONNXRuntime session is not set.")
        input_dict = dict(zip(self.get_input_names(), input_content))
        try:
            return self.session.run(self.get_output_names(), input_dict)
        except Exception as e:
            print(e)
            raise ONNXRuntimeError("ONNXRuntime inferece failed.") from e

    def get_input_names(
        self,
    ):
        return [v.name for v in self.session.get_inputs()]

    def get_output_names(
        self,
    ):
        return [v.name for v in self.session.get_outputs()]

    def get_character_list(self, key: str = "character"):
        return self.meta_dict[key].splitlines()

    def have_key(self, key: str = "character") -> bool:
        self.meta_dict = self.session.get_modelmeta().custom_metadata_map
        if key in self.meta_dict.keys():
            return True
        return False

    @staticmethod
    def _verify_model(model_path):
        model_path = Path(model_path)
        if not model_path.exists():
            raise FileNotFoundError(f"{model_path} does not exists.")
        if not model_path.is_file():
            raise FileExistsError(f"{model_path} is not a file.")

    def set_session(self, intra_op_num_threads: int):
        self.intra_op_num_threads = intra_op_num_threads
        sess_opt = ort.SessionOptions()
        sess_opt.intra_op_num_threads = intra_op_num_threads
        sess_opt.log_severity_level = 4
        sess_opt.enable_cpu_mem_arena = False
        sess_opt.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
        sess_opt.execution_mode = ort.ExecutionMode.ORT_SEQUENTIAL
        EP_list = []
        cuda_ep = "CUDAExecutionProvider"
        cpu_ep = "CPUExecutionProvider"
        if ort.get_device() == "GPU" and cuda_ep in ort.get_available_providers():
            if self.device_id is None:
                device_id = 0
            else:
                device_id = self.device_id
            cuda_provider_options = {
                "device_id": str(device_id),
                "arena_extend_strategy": "kNextPowerOfTwo",
                "cudnn_conv_algo_search": "EXHAUSTIVE",
                "do_copy_in_default_stream": "true",
            }
            EP_list.append((cuda_ep, cuda_provider_options))
            if self.verbose:
                logger.info(f"using onnxruntime-gpu with device_id: {device_id}")
        else:
            cpu_provider_options = {
                "arena_extend_strategy": "kSameAsRequested",
            }
            EP_list.append((cpu_ep, cpu_provider_options))
            if self.verbose:
                logger.info(
                    "using onnxruntime-cpu, if you want to use GPU, please install onnxruntime-gpu"
                )

        self._verify_model(self.onnx_model_path)
        self.session = ort.InferenceSession(
            self.onnx_model_path, sess_options=sess_opt, providers=EP_list
        )

    def get_config(self) -> Config:
        data = {
            "runtime": {
                "@runtimes": "ort",
                "intra_op_num_threads": self.intra_op_num_threads,
            }
        }
        return Config(data)

    def save(self, save_dir: str = "runtime") -> None:
        import shutil
        from shutil import SameFileError

        save_dir = Path(save_dir)
        if not save_dir.exists():
            save_dir.mkdir(parents=True)
        model_file = save_dir / "model.onnx"
        try:
            _ = shutil.copy(self.onnx_model_path, model_file)
        except SameFileError:
            pass
        except Exception as e:
            raise ONNXRuntimeError(f"Failed to save model to {model_file}.") from e
        config = self.get_config()
        config.to_disk(Path(save_dir) / "config.cfg")

    def load(self, save_dir: str = "runtime", **kwargs) -> "ORT":
        save_dir = Path(save_dir)
        config = Config().from_disk(save_dir / "config.cfg")
        ort: "ORT" = registry.resolve(config=config)["runtime"]
        ort.onnx_model_path = save_dir / "model.onnx"
        ort.set_session(ort.intra_op_num_threads)
        return ort

    def from_checkpoint(
        self,
        checkpoint_dir: str,
        device_id: Optional[int] = None,
        intra_op_num_threads: int = 2,
        verbose: bool = True,
    ) -> "ORT":
        model_path = Path(checkpoint_dir) / "model.onnx"
        self.onnx_model_path = model_path
        self.device_id = device_id
        self.intra_op_num_threads = intra_op_num_threads
        self.verbose = verbose
        self.set_session(intra_op_num_threads)
        return self


class ONNXRuntimeError(Exception):
    pass



from __future__ import annotations

from pathlib import Path
from typing import Any, List

import numpy as np
from pydantic import ConfigDict, Field
import torch
from typing_extensions import Self

from fasr.config import registry
from fasr.data import AudioToken, AudioTokenList
from fasr.model import ASRModel


def _result_to_text(item: Any) -> str:
    if isinstance(item, str):
        return item.strip()
    if isinstance(item, dict):
        text = item.get("text")
        return "" if text is None else str(text).strip()
    text_attr = getattr(item, "text", None)
    if text_attr is not None:
        return str(text_attr).strip()
    return str(item).strip()


@registry.asr_models.register("fun_asr_nano")
class FunASRNanoForASR(ASRModel):
    """Fun-ASR-Nano 封装（无时间戳）。"""

    checkpoint: str = "FunAudioLLM/Fun-ASR-Nano-2512"
    endpoint: str = "modelscope"

    fun_model: Any | None = Field(default=None, exclude=True)
    language: str = "中文"
    itn: bool = True
    batch_size: int = 1

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    def transcribe(
        self,
        batch: List[np.ndarray | torch.Tensor],
        **kwargs,
    ) -> List[AudioTokenList]:
        if self.fun_model is None:
            raise RuntimeError("Call from_checkpoint before transcribe")
        sample_rate = int(kwargs.get("sample_rate", 16000))
        hotwords = kwargs.get("hotwords", None)
        language = kwargs.get("language", self.language)
        itn = kwargs.get("itn", self.itn)
        batch_size = int(kwargs.get("batch_size", self.batch_size))

        inputs: list[tuple[np.ndarray, int]] = []
        for waveform in batch:
            if isinstance(waveform, torch.Tensor):
                wav = waveform.detach().cpu().float().numpy()
            else:
                wav = np.asarray(waveform)
            wav = wav.squeeze()
            inputs.append((wav, sample_rate))

        results = self.fun_model.generate(
            input=inputs,
            cache={},
            batch_size=batch_size,
            hotwords=hotwords,
            language=language,
            itn=itn,
        )
        out: list[AudioTokenList] = []
        for item in results:
            text = _result_to_text(item)
            # Fun-ASR-Nano 当前封装按整段文本返回，不提供词级时间戳
            out.append(AudioTokenList([AudioToken(text=text, start_ms=1, end_ms=2)]))
        return out

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        device: str = "cuda:0",
        trust_remote_code: bool = True,
        language: str = "中文",
        itn: bool = True,
        batch_size: int = 1,
        **kwargs,
    ) -> Self:
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")

        from funasr import AutoModel

        self.fun_model = AutoModel(
            model=str(checkpoint_dir),
            trust_remote_code=trust_remote_code,
            device=device,
            **kwargs,
        )
        self.language = language
        self.itn = itn
        self.batch_size = batch_size
        return self

    def get_config(self):
        raise NotImplementedError

    def load(self, save_dir, **kwargs):
        raise NotImplementedError

    def save(self, save_dir, **kwargs):
        raise NotImplementedError



from __future__ import annotations

from pathlib import Path
from typing import Any

from pydantic import ConfigDict, Field
import torch
from typing_extensions import Self

from fasr.config import registry
from fasr.data import Waveform
from fasr.model import LIDModel


@registry.lid_models.register("firered")
class FireRedForLID(LIDModel):
    checkpoint: str = "FireRedTeam/FireRedLID"
    endpoint: str = "modelscope"

    firered: Any | None = Field(default=None, exclude=True)
    use_gpu: bool = True
    use_half: bool = False
    max_chunk_seconds: float = 60.0

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    def identify(self, waveform: Waveform) -> str:
        wav = waveform.resample(16000) if waveform.sample_rate != 16000 else waveform
        max_chunk_samples = int(self.max_chunk_seconds * wav.sample_rate)
        if max_chunk_samples <= 0:
            max_chunk_samples = wav.data.shape[-1]
        min_chunk_samples = int(5 * wav.sample_rate)

        lang_votes: dict[str, tuple[int, float]] = {}
        pending_ranges: list[tuple[int, int]] = [
            (start, min(start + max_chunk_samples, wav.data.shape[-1]))
            for start in range(0, wav.data.shape[-1], max_chunk_samples)
        ]
        chunk_idx = 0
        while pending_ranges:
            start, end = pending_ranges.pop(0)
            chunk = Waveform(data=wav.data[start:end], sample_rate=wav.sample_rate)
            batch_wav = [(chunk.sample_rate, chunk.to_int16_pcm())]
            uttids = [f"utt{chunk_idx}"]
            chunk_idx += 1
            try:
                results = self.firered.process(uttids, batch_wav)
            except torch.OutOfMemoryError:
                if self.use_gpu:
                    if torch.cuda.is_available():
                        torch.cuda.empty_cache()
                    # 优先切更小分块继续在 GPU 上跑，避免整次任务降级到 CPU
                    if end - start > min_chunk_samples:
                        mid = start + (end - start) // 2
                        pending_ranges.insert(0, (mid, end))
                        pending_ranges.insert(0, (start, mid))
                        continue
                    # 当前小分块仍 OOM，仅该分块临时回退到 CPU
                    self.firered.config.use_gpu = False
                    self.firered.model = self.firered.model.cpu()
                    results = self.firered.process(uttids, batch_wav)
                    # 尝试恢复 GPU，供后续分块继续使用
                    try:
                        self.firered.model = self.firered.model.cuda()
                        if self.use_half:
                            self.firered.model.half()
                        self.firered.config.use_gpu = True
                    except torch.OutOfMemoryError:
                        self.firered.config.use_gpu = False
                        self.use_gpu = False
                else:
                    raise
            for item in results:
                lang = (item.get("lang") or "").strip()
                if not lang:
                    continue
                conf = float(item.get("confidence") or 0.0)
                cnt, conf_sum = lang_votes.get(lang, (0, 0.0))
                lang_votes[lang] = (cnt + 1, conf_sum + conf)
        if lang_votes:
            best_lang = max(lang_votes.items(), key=lambda x: (x[1][0], x[1][1]))[0]
            return best_lang
        return ""

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        use_gpu: bool = True,
        use_half: bool = False,
        max_chunk_seconds: float = 60.0,
        **kwargs,
    ) -> Self:
        from fasr_lid_firered.fireredlid.lid import FireRedLid, FireRedLidConfig

        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        config = FireRedLidConfig(use_gpu=use_gpu, use_half=use_half)
        self.firered = FireRedLid.from_pretrained(str(checkpoint_dir), config=config)
        self.use_gpu = use_gpu
        self.use_half = use_half
        self.max_chunk_seconds = max_chunk_seconds
        return self

    def get_config(self):
        raise NotImplementedError

    def load(self, save_dir, **kwargs):
        raise NotImplementedError

    def save(self, save_dir, **kwargs):
        raise NotImplementedError



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
import math
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple

import numpy as np
from pydantic import ConfigDict, Field
import torch
from typing_extensions import Self

from fasr.config import registry
from fasr.data import Audio, AudioChannel, AudioList, AudioSpan, AudioSpanList, Waveform
from fasr.model import VADModel
from fasr.utils import read_yaml

from ..preprocessors.waveform.fbank_extractor import FbankExtractor
from ..runtimes.ort import ORT

DEFAULT_CHECKPOINT_DIR = Path(__file__).parent.parent / "asset" / "fsmn-vad"


@registry.vad_models.register("fsmn")
class FSMNForVAD(VADModel):
    """FSMN VAD模型实现

    FSMN (Feedforward Sequential Memory Networks) VAD是一个基于神经网络的语音活动检测模型。
    使用ONNX Runtime进行推理，支持实时和离线VAD。
    """

    # 模型相关属性
    config: Optional[dict] = Field(default_factory=dict, description="模型配置字典")
    audio_preprocessor: Optional[FbankExtractor] = Field(
        default=None, description="音频预处理器"
    )
    encoder: Optional[Any] = Field(default=None, description="FSMN编码器")
    segment_predictor: Optional[Any] = Field(default=None, description="分段预测器")
    sample_rate: int = Field(default=16000, description="音频采样率")

    # VAD配置参数
    detect_mode: int = Field(default=1, description="检测模式")
    snr_mode: int = Field(default=0, description="信噪比模式")
    max_end_silence_time: int = Field(
        default=800, ge=0, description="最大结束静音时间(毫秒)"
    )
    max_start_silence_time: int = Field(
        default=3000, ge=0, description="最大开始静音时间(毫秒)"
    )
    do_start_point_detection: bool = Field(
        default=True, description="是否进行起始点检测"
    )
    do_end_point_detection: bool = Field(default=True, description="是否进行结束点检测")
    window_size_ms: int = Field(default=200, ge=0, description="窗口大小(毫秒)")
    sil_to_speech_time_thres: int = Field(
        default=150, ge=0, description="静音到语音时间阈值(毫秒)"
    )
    speech_to_sil_time_thres: int = Field(
        default=150, ge=0, description="语音到静音时间阈值(毫秒)"
    )
    speech_2_noise_ratio: float = Field(default=1.0, ge=0.0, description="语音噪声比")
    do_extend: int = Field(default=1, description="是否扩展")
    lookback_time_start_point: int = Field(
        default=200, ge=0, description="起始点回看时间(毫秒)"
    )
    lookahead_time_end_point: int = Field(
        default=100, ge=0, description="结束点前瞻时间(毫秒)"
    )
    max_single_segment_time: int = Field(
        default=60000, ge=0, description="最大单段时长(毫秒)"
    )
    snr_thres: float = Field(default=-100.0, description="信噪比阈值")
    noise_frame_num_used_for_snr: int = Field(
        default=100, ge=0, description="用于SNR计算的噪声帧数"
    )
    decibel_thres: float = Field(default=-100.0, description="分贝阈值")
    speech_noise_thres: float = Field(
        default=0.6, ge=0.0, le=1.0, description="语音噪声阈值"
    )
    fe_prior_thres: float = Field(default=1e-4, ge=0.0, description="前端先验阈值")
    silence_pdf_num: int = Field(default=1, ge=0, description="静音PDF数量")
    sil_pdf_ids: List[int] = Field(default=[0], description="静音PDF ID列表")
    speech_noise_thresh_low: float = Field(default=-0.1, description="语音噪声阈值下限")
    speech_noise_thresh_high: float = Field(default=0.3, description="语音噪声阈值上限")
    output_frame_probs: bool = Field(default=False, description="是否输出帧概率")
    frame_in_ms: int = Field(default=10, ge=0, description="帧间隔(毫秒)")
    frame_length_ms: int = Field(default=25, ge=0, description="帧长度(毫秒)")

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    def _update_config_dict(self):
        """更新配置字典"""
        if self.config is None:
            self.config = {}

        # 更新model_conf部分
        if "model_conf" not in self.config:
            self.config["model_conf"] = {}

        self.config["model_conf"].update(
            {
                "detect_mode": self.detect_mode,
                "snr_mode": self.snr_mode,
                "max_end_silence_time": self.max_end_silence_time,
                "max_start_silence_time": self.max_start_silence_time,
                "do_start_point_detection": self.do_start_point_detection,
                "do_end_point_detection": self.do_end_point_detection,
                "window_size_ms": self.window_size_ms,
                "sil_to_speech_time_thres": self.sil_to_speech_time_thres,
                "speech_to_sil_time_thres": self.speech_to_sil_time_thres,
                "speech_2_noise_ratio": self.speech_2_noise_ratio,
                "do_extend": self.do_extend,
                "lookback_time_start_point": self.lookback_time_start_point,
                "lookahead_time_end_point": self.lookahead_time_end_point,
                "max_single_segment_time": self.max_single_segment_time,
                "snr_thres": self.snr_thres,
                "noise_frame_num_used_for_snr": self.noise_frame_num_used_for_snr,
                "decibel_thres": self.decibel_thres,
                "speech_noise_thres": self.speech_noise_thres,
                "fe_prior_thres": self.fe_prior_thres,
                "silence_pdf_num": self.silence_pdf_num,
                "sil_pdf_ids": self.sil_pdf_ids,
                "speech_noise_thresh_low": self.speech_noise_thresh_low,
                "speech_noise_thresh_high": self.speech_noise_thresh_high,
                "output_frame_probs": self.output_frame_probs,
                "frame_in_ms": self.frame_in_ms,
                "frame_length_ms": self.frame_length_ms,
            }
        )

    def update_config(self, **kwargs):
        """动态更新VAD配置参数

        Args:
            **kwargs: 要更新的配置参数
        """
        # 更新实例属性
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

        # 更新配置字典
        self._update_config_dict()

        print(f"✅ FSMN VAD配置已更新: {kwargs}")

    def get_config(self) -> dict:
        """获取当前VAD配置

        Returns:
            dict: 当前配置字典
        """
        return {
            "detect_mode": self.detect_mode,
            "snr_mode": self.snr_mode,
            "max_end_silence_time": self.max_end_silence_time,
            "max_start_silence_time": self.max_start_silence_time,
            "do_start_point_detection": self.do_start_point_detection,
            "do_end_point_detection": self.do_end_point_detection,
            "window_size_ms": self.window_size_ms,
            "sil_to_speech_time_thres": self.sil_to_speech_time_thres,
            "speech_to_sil_time_thres": self.speech_to_sil_time_thres,
            "speech_2_noise_ratio": self.speech_2_noise_ratio,
            "do_extend": self.do_extend,
            "lookback_time_start_point": self.lookback_time_start_point,
            "lookahead_time_end_point": self.lookahead_time_end_point,
            "max_single_segment_time": self.max_single_segment_time,
            "snr_thres": self.snr_thres,
            "noise_frame_num_used_for_snr": self.noise_frame_num_used_for_snr,
            "decibel_thres": self.decibel_thres,
            "speech_noise_thres": self.speech_noise_thres,
            "fe_prior_thres": self.fe_prior_thres,
            "silence_pdf_num": self.silence_pdf_num,
            "sil_pdf_ids": self.sil_pdf_ids,
            "speech_noise_thresh_low": self.speech_noise_thresh_low,
            "speech_noise_thresh_high": self.speech_noise_thresh_high,
            "output_frame_probs": self.output_frame_probs,
            "frame_in_ms": self.frame_in_ms,
            "frame_length_ms": self.frame_length_ms,
        }

    def detect(self, waveform: Waveform) -> List[AudioSpan]:
        if waveform.sample_rate != self.sample_rate:
            waveform = waveform.resample(self.sample_rate)
        waveform = self.audio_preprocessor.process_waveform(waveform=waveform)
        steps = self.encoder.encode_waveform(waveform=waveform)
        predictor = SegmentPredictor.from_config(
            self.config["model_conf"]
        )  # 解决多线程问题
        segments = predictor.predict_steps(steps=steps)
        for segment in segments:
            segment.waveform = waveform.select_by_ms(
                start=segment.start_ms, end=segment.end_ms
            )
        return segments

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path = None,
        device_id: str | int | None = None,
        num_threads: int = 2,
        compile: bool = False,
        **kwargs,
    ) -> Self:
        """从检查点加载模型

        Args:
            checkpoint_dir: 检查点目录路径
            device_id: 设备ID
            num_threads: 线程数
            compile: 是否编译
            **kwargs: 其他参数

        Returns:
            Self: 模型实例
        """
        # 确定检查点目录
        checkpoint_dir = (
            kwargs.get("checkpoint_dir", checkpoint_dir) or DEFAULT_CHECKPOINT_DIR
        )
        config_path = Path(checkpoint_dir) / "config.yaml"

        if config_path.exists():
            file_config = read_yaml(config_path)
        else:
            # 使用默认配置
            file_config = {}

        # 合并配置：优先使用实例配置，然后使用文件配置
        if self.config is None:
            self.config = {}

        # 先加载文件配置
        self.config.update(file_config)

        # 然后覆盖为实例配置（如果已设置）
        self._update_config_dict()

        # 加载音频预处理器
        audio_preprocessor = FbankExtractor().from_checkpoint(
            checkpoint_dir=checkpoint_dir, compile=compile, sample_rate=self.sample_rate
        )
        self.audio_preprocessor = audio_preprocessor

        # 加载编码器
        encoder = Encoder().from_checkpoint(
            checkpoint_dir=checkpoint_dir,
            device_id=device_id,
            intra_op_num_threads=num_threads,
        )
        self.encoder = encoder

        return self


@dataclass
class Encoder:
    num_threads: int = 2
    n_fsmn_layers: int = None
    proj_dim: int = None
    lorder: int = None
    runtime: ORT | None = None

    def encode_waveform(self, waveform: Waveform) -> Tuple[np.ndarray, np.ndarray]:
        steps = []
        in_cache = self.prepare_cache()
        data = waveform.data
        feats = waveform.feats
        feats = feats[None, :].astype(np.float32)
        feats_len = feats.shape[1]
        data = np.array(data)[None, :].astype(np.float32)
        t_offset = 0
        step = int(min(feats_len, 6000))
        for t_offset in range(0, int(feats_len), min(step, feats_len - t_offset)):
            if t_offset + step >= feats_len - 1:
                step = feats_len - t_offset
            feats_package = feats[:, t_offset : int(t_offset + step), :]
            data_package = data[
                :,
                t_offset * 160 : min(
                    data.shape[-1], (int(t_offset + step) - 1) * 160 + 400
                ),
            ]
            inputs = [feats_package]
            inputs.extend(in_cache)
            # cache [cache1, cache2, cache3, cache4]
            outputs = self.runtime.run(inputs)
            scores, out_caches = outputs[0], outputs[1:]
            steps.append(
                Waveform(
                    data=data_package, sample_rate=waveform.sample_rate, scores=scores
                )
            )
            in_cache = out_caches
        return steps

    @lru_cache(maxsize=1)
    def prepare_cache(self):
        """Prepare cache for FSMN model.

        Returns:
            List: List of cache for FSMN model. shape = (n_layers, proj_dim, lorder - 1, 1)
        """
        in_cache = []
        for i in range(self.n_fsmn_layers):
            cache = np.zeros((1, self.proj_dim, self.lorder - 1, 1)).astype(np.float32)
            in_cache.append(cache)
        return in_cache

    def from_checkpoint(
        self,
        checkpoint_dir: str = "checkpoints/iic/speech_fsmn_vad_zh-cn-16k-common-pytorch",
        device_id: Optional[int] = None,
        intra_op_num_threads: int = 2,
        num_threads: int = 2,
    ) -> Self:
        config = read_yaml(Path(checkpoint_dir, "config.yaml"))
        encoder_config = config["encoder_conf"]
        self.n_fsmn_layers = encoder_config["fsmn_layers"]
        self.proj_dim = encoder_config["proj_dim"]
        self.lorder = encoder_config["lorder"]
        runtime = ORT().from_checkpoint(
            checkpoint_dir=checkpoint_dir,
            device_id=device_id,
            intra_op_num_threads=intra_op_num_threads,
        )
        self.runtime = runtime
        self.num_threads = num_threads
        return self

    def __hash__(self):
        return hash(f"{self.n_fsmn_layers}-{self.proj_dim}-{self.lorder}")


class VadStateMachine(Enum):
    kVadInStateStartPointNotDetected = 1
    kVadInStateInSpeechSegment = 2
    kVadInStateEndPointDetected = 3


class FrameState(Enum):
    kFrameStateInvalid = -1
    kFrameStateSpeech = 1
    kFrameStateSil = 0


# final voice/unvoice state per frame
class AudioChangeState(Enum):
    kChangeStateSpeech2Speech = 0
    kChangeStateSpeech2Sil = 1
    kChangeStateSil2Sil = 2
    kChangeStateSil2Speech = 3
    kChangeStateNoBegin = 4
    kChangeStateInvalid = 5


class VadDetectMode(Enum):
    kVadSingleUtteranceDetectMode = 0
    kVadMutipleUtteranceDetectMode = 1


class VADXOptions:
    def __init__(
        self,
        sample_rate: int = 16000,
        detect_mode: int = VadDetectMode.kVadMutipleUtteranceDetectMode.value,
        snr_mode: int = 0,
        max_end_silence_time: int = 800,
        max_start_silence_time: int = 3000,
        do_start_point_detection: bool = True,
        do_end_point_detection: bool = True,
        window_size_ms: int = 200,
        sil_to_speech_time_thres: int = 150,
        speech_to_sil_time_thres: int = 150,
        speech_2_noise_ratio: float = 1.0,
        do_extend: int = 1,
        lookback_time_start_point: int = 200,
        lookahead_time_end_point: int = 100,
        max_single_segment_time: int = 60000,
        nn_eval_block_size: int = 8,
        dcd_block_size: int = 4,
        snr_thres: int = -100.0,
        noise_frame_num_used_for_snr: int = 100,
        decibel_thres: int = -100.0,
        speech_noise_thres: float = 0.6,
        fe_prior_thres: float = 1e-4,
        silence_pdf_num: int = 1,
        sil_pdf_ids: List[int] = [0],
        speech_noise_thresh_low: float = -0.1,
        speech_noise_thresh_high: float = 0.3,
        output_frame_probs: bool = False,
        frame_in_ms: int = 10,
        frame_length_ms: int = 25,
    ):
        self.sample_rate = sample_rate
        self.detect_mode = detect_mode
        self.snr_mode = snr_mode
        self.max_end_silence_time = max_end_silence_time
        self.max_start_silence_time = max_start_silence_time
        self.do_start_point_detection = do_start_point_detection
        self.do_end_point_detection = do_end_point_detection
        self.window_size_ms = window_size_ms
        self.sil_to_speech_time_thres = sil_to_speech_time_thres
        self.speech_to_sil_time_thres = speech_to_sil_time_thres
        self.speech_2_noise_ratio = speech_2_noise_ratio
        self.do_extend = do_extend
        self.lookback_time_start_point = lookback_time_start_point
        self.lookahead_time_end_point = lookahead_time_end_point
        self.max_single_segment_time = max_single_segment_time
        self.nn_eval_block_size = nn_eval_block_size
        self.dcd_block_size = dcd_block_size
        self.snr_thres = snr_thres
        self.noise_frame_num_used_for_snr = noise_frame_num_used_for_snr
        self.decibel_thres = decibel_thres
        self.speech_noise_thres = speech_noise_thres
        self.fe_prior_thres = fe_prior_thres
        self.silence_pdf_num = silence_pdf_num
        self.sil_pdf_ids = sil_pdf_ids
        self.speech_noise_thresh_low = speech_noise_thresh_low
        self.speech_noise_thresh_high = speech_noise_thresh_high
        self.output_frame_probs = output_frame_probs
        self.frame_in_ms = frame_in_ms
        self.frame_length_ms = frame_length_ms


class E2EVadSpeechBufWithDoa(object):
    def __init__(self):
        self.start_ms = 0
        self.end_ms = 0
        self.buffer = []
        self.contain_seg_start_point = False
        self.contain_seg_end_point = False
        self.doa = 0

    def Reset(self):
        self.start_ms = 0
        self.end_ms = 0
        self.buffer = []
        self.contain_seg_start_point = False
        self.contain_seg_end_point = False
        self.doa = 0


class E2EVadFrameProb(object):
    def __init__(self):
        self.noise_prob = 0.0
        self.speech_prob = 0.0
        self.score = 0.0
        self.frame_id = 0
        self.frm_state = 0


class WindowDetector(object):
    def __init__(
        self,
        window_size_ms: int,
        sil_to_speech_time: int,
        speech_to_sil_time: int,
        frame_size_ms: int,
    ):
        self.window_size_ms = window_size_ms
        self.sil_to_speech_time = sil_to_speech_time
        self.speech_to_sil_time = speech_to_sil_time
        self.frame_size_ms = frame_size_ms

        self.win_size_frame = int(window_size_ms / frame_size_ms)
        self.win_sum = 0
        self.win_state = [0] * self.win_size_frame  # 初始化窗

        self.cur_win_pos = 0
        self.pre_frame_state = FrameState.kFrameStateSil
        self.cur_frame_state = FrameState.kFrameStateSil
        self.sil_to_speech_frmcnt_thres = int(sil_to_speech_time / frame_size_ms)
        self.speech_to_sil_frmcnt_thres = int(speech_to_sil_time / frame_size_ms)

        self.voice_last_frame_count = 0
        self.noise_last_frame_count = 0
        self.hydre_frame_count = 0

    def Reset(self) -> None:
        self.cur_win_pos = 0
        self.win_sum = 0
        self.win_state = [0] * self.win_size_frame
        self.pre_frame_state = FrameState.kFrameStateSil
        self.cur_frame_state = FrameState.kFrameStateSil
        self.voice_last_frame_count = 0
        self.noise_last_frame_count = 0
        self.hydre_frame_count = 0

    def GetWinSize(self) -> int:
        return int(self.win_size_frame)

    def DetectOneFrame(
        self, frameState: FrameState, frame_count: int
    ) -> AudioChangeState:
        cur_frame_state = FrameState.kFrameStateSil
        if frameState == FrameState.kFrameStateSpeech:
            cur_frame_state = 1
        elif frameState == FrameState.kFrameStateSil:
            cur_frame_state = 0
        else:
            return AudioChangeState.kChangeStateInvalid
        self.win_sum -= self.win_state[self.cur_win_pos]
        self.win_sum += cur_frame_state
        self.win_state[self.cur_win_pos] = cur_frame_state
        self.cur_win_pos = (self.cur_win_pos + 1) % self.win_size_frame

        if (
            self.pre_frame_state == FrameState.kFrameStateSil
            and self.win_sum >= self.sil_to_speech_frmcnt_thres
        ):
            self.pre_frame_state = FrameState.kFrameStateSpeech
            return AudioChangeState.kChangeStateSil2Speech

        if (
            self.pre_frame_state == FrameState.kFrameStateSpeech
            and self.win_sum <= self.speech_to_sil_frmcnt_thres
        ):
            self.pre_frame_state = FrameState.kFrameStateSil
            return AudioChangeState.kChangeStateSpeech2Sil

        if self.pre_frame_state == FrameState.kFrameStateSil:
            return AudioChangeState.kChangeStateSil2Sil
        if self.pre_frame_state == FrameState.kFrameStateSpeech:
            return AudioChangeState.kChangeStateSpeech2Speech
        return AudioChangeState.kChangeStateInvalid

    def FrameSizeMs(self) -> int:
        return int(self.frame_size_ms)


class SegmentPredictor:
    """
    Author: Speech Lab of DAMO Academy, Alibaba Group
    Deep-FSMN for Large Vocabulary Continuous Speech Recognition
    https://arxiv.org/abs/1803.05030
    """

    def __init__(self, vad_post_args: Dict[str, Any]):
        super().__init__()
        self.vad_opts = VADXOptions(**vad_post_args)
        self.windows_detector = WindowDetector(
            self.vad_opts.window_size_ms,
            self.vad_opts.sil_to_speech_time_thres,
            self.vad_opts.speech_to_sil_time_thres,
            self.vad_opts.frame_in_ms,
        )
        self.is_final = False
        self.data_buf_start_frame = 0
        self.frm_cnt = 0
        self.latest_confirmed_speech_frame = 0
        self.lastest_confirmed_silence_frame = -1
        self.continous_silence_frame_count = 0
        self.vad_state_machine = VadStateMachine.kVadInStateStartPointNotDetected
        self.confirmed_start_frame = -1
        self.confirmed_end_frame = -1
        self.number_end_time_detected = 0
        self.sil_frame = 0
        self.sil_pdf_ids = self.vad_opts.sil_pdf_ids
        self.noise_average_decibel = -100.0
        self.pre_end_silence_detected = False
        self.next_seg = True

        self.output_data_buf = []
        self.output_data_buf_offset = 0
        self.frame_probs = []
        self.max_end_sil_frame_cnt_thresh = (
            self.vad_opts.max_end_silence_time - self.vad_opts.speech_to_sil_time_thres
        )
        self.speech_noise_thres = self.vad_opts.speech_noise_thres
        self.scores = None
        self.idx_pre_chunk = 0
        self.max_time_out = False
        self.decibel = []
        self.data_buf_size = 0
        self.data_buf_all_size = 0
        self.waveform = None
        self.ResetDetection()

        self.timer_data = {}

    def AllResetDetection(self):
        self.is_final = False
        self.data_buf_start_frame = 0
        self.frm_cnt = 0
        self.latest_confirmed_speech_frame = 0
        self.lastest_confirmed_silence_frame = -1
        self.continous_silence_frame_count = 0
        self.vad_state_machine = VadStateMachine.kVadInStateStartPointNotDetected
        self.confirmed_start_frame = -1
        self.confirmed_end_frame = -1
        self.number_end_time_detected = 0
        self.sil_frame = 0
        self.sil_pdf_ids = self.vad_opts.sil_pdf_ids
        self.noise_average_decibel = -100.0
        self.pre_end_silence_detected = False
        self.next_seg = True

        self.output_data_buf = []
        self.output_data_buf_offset = 0
        self.frame_probs = []
        self.max_end_sil_frame_cnt_thresh = (
            self.vad_opts.max_end_silence_time - self.vad_opts.speech_to_sil_time_thres
        )
        self.speech_noise_thres = self.vad_opts.speech_noise_thres
        self.scores = None
        self.idx_pre_chunk = 0
        self.max_time_out = False
        self.decibel = []
        self.data_buf_size = 0
        self.data_buf_all_size = 0
        self.waveform = None
        self.ResetDetection()

    def ResetDetection(self):
        self.continous_silence_frame_count = 0
        self.latest_confirmed_speech_frame = 0
        self.lastest_confirmed_silence_frame = -1
        self.confirmed_start_frame = -1
        self.confirmed_end_frame = -1
        self.vad_state_machine = VadStateMachine.kVadInStateStartPointNotDetected
        self.windows_detector.Reset()
        self.sil_frame = 0
        self.frame_probs = []

    def ComputeDecibel(self) -> None:
        frame_sample_length = int(
            self.vad_opts.frame_length_ms * self.vad_opts.sample_rate / 1000
        )
        frame_shift_length = int(
            self.vad_opts.frame_in_ms * self.vad_opts.sample_rate / 1000
        )
        if self.data_buf_all_size == 0:
            self.data_buf_all_size = len(self.waveform[0])
            self.data_buf_size = self.data_buf_all_size
        else:
            self.data_buf_all_size += len(self.waveform[0])
        for offset in range(
            0, self.waveform.shape[1] - frame_sample_length + 1, frame_shift_length
        ):
            self.decibel.append(
                10
                * math.log10(
                    np.square(
                        (self.waveform[0][offset : offset + frame_sample_length])
                    ).sum()
                    + 0.000001
                )
            )

    def compute_decibel_fast(self) -> None:
        """使用pytorch重新实现了计算分贝过程，并基于torch.compile加速计算"""
        if self.data_buf_all_size == 0:
            self.data_buf_all_size = len(self.waveform[0])
            self.data_buf_size = self.data_buf_all_size
        else:
            self.data_buf_all_size += len(self.waveform[0])
        self.decibel.extend(
            compute_audio_db(
                self.waveform,
                sr=self.vad_opts.sample_rate,
                frame_length_ms=self.vad_opts.frame_length_ms,
                frame_shift_ms=self.vad_opts.frame_in_ms,
            )
        )

    def ComputeScores(self, scores: np.ndarray) -> None:
        self.vad_opts.nn_eval_block_size = scores.shape[1]  # num of frames
        self.frm_cnt += scores.shape[1]  # count total frames
        self.scores = scores

    def PopDataBufTillFrame(self, frame_idx: int) -> None:  # need check again
        while self.data_buf_start_frame < frame_idx:
            if self.data_buf_size >= int(
                self.vad_opts.frame_in_ms * self.vad_opts.sample_rate / 1000
            ):
                self.data_buf_start_frame += 1
                self.data_buf_size = (
                    self.data_buf_all_size
                    - self.data_buf_start_frame
                    * int(self.vad_opts.frame_in_ms * self.vad_opts.sample_rate / 1000)
                )

    def PopDataToOutputBuf(
        self,
        start_frm: int,
        frm_cnt: int,
        first_frm_is_start_point: bool,
        last_frm_is_end_point: bool,
        end_point_is_sent_end: bool,
    ) -> None:
        self.PopDataBufTillFrame(start_frm)
        expected_sample_number = int(
            frm_cnt * self.vad_opts.sample_rate * self.vad_opts.frame_in_ms / 1000
        )
        if last_frm_is_end_point:
            extra_sample = max(
                0,
                int(
                    self.vad_opts.frame_length_ms * self.vad_opts.sample_rate / 1000
                    - self.vad_opts.sample_rate * self.vad_opts.frame_in_ms / 1000
                ),
            )
            expected_sample_number += int(extra_sample)
        if end_point_is_sent_end:
            expected_sample_number = max(expected_sample_number, self.data_buf_size)
        if self.data_buf_size < expected_sample_number:
            print("error in calling pop data_buf\n")

        if len(self.output_data_buf) == 0 or first_frm_is_start_point:
            self.output_data_buf.append(E2EVadSpeechBufWithDoa())
            self.output_data_buf[-1].Reset()
            self.output_data_buf[-1].start_ms = start_frm * self.vad_opts.frame_in_ms
            self.output_data_buf[-1].end_ms = self.output_data_buf[-1].start_ms
            self.output_data_buf[-1].doa = 0
        cur_seg = self.output_data_buf[-1]
        if cur_seg.end_ms != start_frm * self.vad_opts.frame_in_ms:
            print("warning\n")
        out_pos = len(cur_seg.buffer)  # cur_seg.buff现在没做任何操作
        data_to_pop = 0
        if end_point_is_sent_end:
            data_to_pop = expected_sample_number
        else:
            data_to_pop = int(
                frm_cnt * self.vad_opts.frame_in_ms * self.vad_opts.sample_rate / 1000
            )
        if data_to_pop > self.data_buf_size:
            print("VAD data_to_pop is bigger than self.data_buf_size!!!\n")
            data_to_pop = self.data_buf_size
            expected_sample_number = self.data_buf_size

        cur_seg.doa = 0
        for sample_cpy_out in range(0, data_to_pop):
            # cur_seg.buffer[out_pos ++] = data_buf_.back();
            out_pos += 1
        for sample_cpy_out in range(data_to_pop, expected_sample_number):
            # cur_seg.buffer[out_pos++] = data_buf_.back()
            out_pos += 1
        if cur_seg.end_ms != start_frm * self.vad_opts.frame_in_ms:
            print("Something wrong with the VAD algorithm\n")
        self.data_buf_start_frame += frm_cnt
        cur_seg.end_ms = (start_frm + frm_cnt) * self.vad_opts.frame_in_ms
        if first_frm_is_start_point:
            cur_seg.contain_seg_start_point = True
        if last_frm_is_end_point:
            cur_seg.contain_seg_end_point = True

    def OnSilenceDetected(self, valid_frame: int):
        self.lastest_confirmed_silence_frame = valid_frame
        if self.vad_state_machine == VadStateMachine.kVadInStateStartPointNotDetected:
            self.PopDataBufTillFrame(valid_frame)
        # silence_detected_callback_
        # pass

    def OnVoiceDetected(self, valid_frame: int) -> None:
        self.latest_confirmed_speech_frame = valid_frame
        self.PopDataToOutputBuf(valid_frame, 1, False, False, False)

    def OnVoiceStart(self, start_frame: int, fake_result: bool = False) -> None:
        if self.vad_opts.do_start_point_detection:
            pass
        if self.confirmed_start_frame != -1:
            print("not reset vad properly\n")
        else:
            self.confirmed_start_frame = start_frame

        if (
            not fake_result
            and self.vad_state_machine
            == VadStateMachine.kVadInStateStartPointNotDetected
        ):
            self.PopDataToOutputBuf(self.confirmed_start_frame, 1, True, False, False)

    def OnVoiceEnd(
        self, end_frame: int, fake_result: bool, is_last_frame: bool
    ) -> None:
        for t in range(self.latest_confirmed_speech_frame + 1, end_frame):
            self.OnVoiceDetected(t)
        if self.vad_opts.do_end_point_detection:
            pass
        if self.confirmed_end_frame != -1:
            print("not reset vad properly\n")
        else:
            self.confirmed_end_frame = end_frame
        if not fake_result:
            self.sil_frame = 0
            self.PopDataToOutputBuf(
                self.confirmed_end_frame, 1, False, True, is_last_frame
            )
        self.number_end_time_detected += 1

    def MaybeOnVoiceEndIfLastFrame(
        self, is_final_frame: bool, cur_frm_idx: int
    ) -> None:
        if is_final_frame:
            self.OnVoiceEnd(cur_frm_idx, False, True)
            self.vad_state_machine = VadStateMachine.kVadInStateEndPointDetected

    def GetLatency(self) -> int:
        return int(self.LatencyFrmNumAtStartPoint() * self.vad_opts.frame_in_ms)

    def LatencyFrmNumAtStartPoint(self) -> int:
        vad_latency = self.windows_detector.GetWinSize()
        if self.vad_opts.do_extend:
            vad_latency += int(
                self.vad_opts.lookback_time_start_point / self.vad_opts.frame_in_ms
            )
        return vad_latency

    def GetFrameState(self, t: int) -> FrameState:
        frame_state = FrameState.kFrameStateInvalid
        cur_decibel = self.decibel[t]
        cur_snr = cur_decibel - self.noise_average_decibel
        # for each frame, calc log posterior probability of each state
        if cur_decibel < self.vad_opts.decibel_thres:
            frame_state = FrameState.kFrameStateSil
            self.DetectOneFrame(frame_state, t, False)
            return frame_state

        sum_score = 0.0
        noise_prob = 0.0
        assert len(self.sil_pdf_ids) == self.vad_opts.silence_pdf_num
        if len(self.sil_pdf_ids) > 0:
            assert len(self.scores) == 1  # 只支持batch_size = 1的测试
            sil_pdf_scores = [
                self.scores[0][t - self.idx_pre_chunk][sil_pdf_id]
                for sil_pdf_id in self.sil_pdf_ids
            ]
            sum_score = sum(sil_pdf_scores)
            noise_prob = math.log(sum_score) * self.vad_opts.speech_2_noise_ratio
            total_score = 1.0
            sum_score = total_score - sum_score
        speech_prob = math.log(sum_score)
        if self.vad_opts.output_frame_probs:
            frame_prob = E2EVadFrameProb()
            frame_prob.noise_prob = noise_prob
            frame_prob.speech_prob = speech_prob
            frame_prob.score = sum_score
            frame_prob.frame_id = t
            self.frame_probs.append(frame_prob)
        if math.exp(speech_prob) >= math.exp(noise_prob) + self.speech_noise_thres:
            if (
                cur_snr >= self.vad_opts.snr_thres
                and cur_decibel >= self.vad_opts.decibel_thres
            ):
                frame_state = FrameState.kFrameStateSpeech
            else:
                frame_state = FrameState.kFrameStateSil
        else:
            frame_state = FrameState.kFrameStateSil
            if self.noise_average_decibel < -99.9:
                self.noise_average_decibel = cur_decibel
            else:
                self.noise_average_decibel = (
                    cur_decibel
                    + self.noise_average_decibel
                    * (self.vad_opts.noise_frame_num_used_for_snr - 1)
                ) / self.vad_opts.noise_frame_num_used_for_snr

        return frame_state

    def __call__(
        self,
        score: np.ndarray,
        waveform: np.ndarray,
        is_final: bool = False,
        max_end_sil: int = 800,
        online: bool = False,
    ):
        self.max_end_sil_frame_cnt_thresh = (
            max_end_sil - self.vad_opts.speech_to_sil_time_thres
        )
        self.waveform = waveform  # compute decibel for each frame
        self.ComputeDecibel()  # 计算当前步的分贝值，添加到self.decibel中
        # self.compute_decibel_fast()
        self.ComputeScores(score)
        if not is_final:
            self.DetectCommonFrames()
        else:
            self.DetectLastFrames()
        segments = []
        for batch_num in range(0, score.shape[0]):  # only support batch_size = 1 now
            segment_batch = []
            if len(self.output_data_buf) > 0:
                for i in range(self.output_data_buf_offset, len(self.output_data_buf)):
                    if online:
                        if not self.output_data_buf[i].contain_seg_start_point:
                            continue
                        if (
                            not self.next_seg
                            and not self.output_data_buf[i].contain_seg_end_point
                        ):
                            continue
                        start_ms = (
                            self.output_data_buf[i].start_ms if self.next_seg else -1
                        )
                        if self.output_data_buf[i].contain_seg_end_point:
                            end_ms = self.output_data_buf[i].end_ms
                            self.next_seg = True
                            self.output_data_buf_offset += 1
                        else:
                            end_ms = -1
                            self.next_seg = False
                    else:
                        if not is_final and (
                            not self.output_data_buf[i].contain_seg_start_point
                            or not self.output_data_buf[i].contain_seg_end_point
                        ):
                            continue
                        start_ms = self.output_data_buf[i].start_ms
                        end_ms = self.output_data_buf[i].end_ms
                        self.output_data_buf_offset += 1
                    segment = [start_ms, end_ms]
                    segment_batch.append(segment)

            if segment_batch:
                segments.append(segment_batch)
        if is_final:
            # reset class variables and clear the dict for the next query
            self.AllResetDetection()
        return segments

    def predict(self, audios: AudioList[Audio], **kwargs) -> AudioList[Audio]:
        for audio in audios:
            audio: Audio
            for i in range(len(audio.channels)):
                channel: AudioChannel = audio.channels[i]
                if channel.steps is None:
                    continue
                channel_segments = []
                for j, step in enumerate(channel.steps):
                    step: AudioSpan
                    if j == len(channel.steps) - 1:
                        is_final = True
                    else:
                        is_final = False
                    segments = self(step.scores, step.waveform, is_final)
                    if segments:
                        channel_segments.extend(segments[0])
                channel.segments = AudioSpanList(
                    docs=[
                        AudioSpan(
                            start_ms=seg[0],
                            end_ms=seg[1],
                            waveform=channel.waveform[
                                int(seg[0] * channel.sample_rate / 1000) : int(
                                    seg[1] * channel.sample_rate / 1000
                                )
                            ],
                        )
                        for seg in channel_segments
                    ]
                )
        return audios

    def predict_stream(
        self, channels: Iterable[AudioChannel], **kwargs
    ) -> Iterable[AudioSpan]:
        for channel in channels:
            sample_rate = channel.sample_rate
            channel: AudioChannel
            if channel.steps is None:
                continue
            channel_segments = AudioSpanList()
            for j, step in enumerate(channel.steps):
                step: AudioSpan
                if j == len(channel.steps) - 1:
                    is_final = True
                else:
                    is_final = False
                segments = self(step.scores, step.waveform, is_final)
                if segments:
                    for seg in segments[0]:
                        _seg = AudioSpan(
                            start_ms=seg[0],
                            end_ms=seg[1],
                            waveform=channel.waveform[
                                int(seg[0] * channel.sample_rate / 1000) : int(
                                    seg[1] * channel.sample_rate / 1000
                                )
                            ],
                            sample_rate=sample_rate,
                        )
                        yield _seg
                        channel_segments.append(_seg)
            channel.segments = channel_segments
            # channel.steps = None  # release memory

    def predict_audio(self, audio: Audio) -> Audio:
        for i in range(len(audio.channels)):
            channel: AudioChannel = audio.channels[i]
            channel_segments = []
            if channel.steps is None:
                continue
            for j, step in enumerate(channel.steps):
                step: AudioSpan
                if j == len(channel.steps) - 1:
                    is_final = True
                else:
                    is_final = False
                segments = self(step.scores, step.waveform, is_final)
                if segments:
                    channel_segments.extend(segments[0])
            channel.segments = AudioSpanList(
                docs=[
                    AudioSpan(
                        start_ms=seg[0],
                        end_ms=seg[1],
                        waveform=channel.waveform[
                            int(seg[0] * channel.sample_rate / 1000) : int(
                                seg[1] * channel.sample_rate / 1000
                            )
                        ],
                    )
                    for seg in channel_segments
                ]
            )
        return audio

    def predict_steps(self, steps: List[Waveform]) -> AudioSpanList:
        channel_segments = AudioSpanList()
        for j, waveform in enumerate(steps):
            if j == len(steps) - 1:
                is_final = True
            else:
                is_final = False
            segments = self(waveform.scores, waveform.data, is_final)
            if segments:
                for seg in segments[0]:
                    _seg = AudioSpan(
                        start_ms=seg[0],
                        end_ms=seg[1],
                        is_last=False,
                    )
                    channel_segments.append(_seg)
        if len(channel_segments) > 0:
            channel_segments[-1].is_last = True
        return channel_segments

    def DetectCommonFrames(self) -> int:
        if self.vad_state_machine == VadStateMachine.kVadInStateEndPointDetected:
            return 0
        for i in range(self.vad_opts.nn_eval_block_size - 1, -1, -1):
            frame_state = FrameState.kFrameStateInvalid
            frame_state = self.GetFrameState(self.frm_cnt - 1 - i)
            self.DetectOneFrame(frame_state, self.frm_cnt - 1 - i, False)
        self.idx_pre_chunk += self.scores.shape[1]
        return 0

    def DetectLastFrames(self) -> int:
        if self.vad_state_machine == VadStateMachine.kVadInStateEndPointDetected:
            return 0
        for i in range(self.vad_opts.nn_eval_block_size - 1, -1, -1):
            frame_state = FrameState.kFrameStateInvalid
            frame_state = self.GetFrameState(self.frm_cnt - 1 - i)
            if i != 0:
                self.DetectOneFrame(frame_state, self.frm_cnt - 1 - i, False)
            else:
                self.DetectOneFrame(frame_state, self.frm_cnt - 1, True)

        return 0

    def DetectOneFrame(
        self, cur_frm_state: FrameState, cur_frm_idx: int, is_final_frame: bool
    ) -> None:
        tmp_cur_frm_state = FrameState.kFrameStateInvalid
        if cur_frm_state == FrameState.kFrameStateSpeech:
            if math.fabs(1.0) > self.vad_opts.fe_prior_thres:
                tmp_cur_frm_state = FrameState.kFrameStateSpeech
            else:
                tmp_cur_frm_state = FrameState.kFrameStateSil
        elif cur_frm_state == FrameState.kFrameStateSil:
            tmp_cur_frm_state = FrameState.kFrameStateSil
        state_change = self.windows_detector.DetectOneFrame(
            tmp_cur_frm_state, cur_frm_idx
        )
        frm_shift_in_ms = self.vad_opts.frame_in_ms
        if AudioChangeState.kChangeStateSil2Speech == state_change:
            silence_frame_count = self.continous_silence_frame_count  # noqa
            self.continous_silence_frame_count = 0
            self.pre_end_silence_detected = False
            start_frame = 0
            if (
                self.vad_state_machine
                == VadStateMachine.kVadInStateStartPointNotDetected
            ):
                start_frame = max(
                    self.data_buf_start_frame,
                    cur_frm_idx - self.LatencyFrmNumAtStartPoint(),
                )
                self.OnVoiceStart(start_frame)
                self.vad_state_machine = VadStateMachine.kVadInStateInSpeechSegment
                for t in range(start_frame + 1, cur_frm_idx + 1):
                    self.OnVoiceDetected(t)
            elif self.vad_state_machine == VadStateMachine.kVadInStateInSpeechSegment:
                for t in range(self.latest_confirmed_speech_frame + 1, cur_frm_idx):
                    self.OnVoiceDetected(t)
                if (
                    cur_frm_idx - self.confirmed_start_frame + 1
                    > self.vad_opts.max_single_segment_time / frm_shift_in_ms
                ):
                    self.OnVoiceEnd(cur_frm_idx, False, False)
                    self.vad_state_machine = VadStateMachine.kVadInStateEndPointDetected
                elif not is_final_frame:
                    self.OnVoiceDetected(cur_frm_idx)
                else:
                    self.MaybeOnVoiceEndIfLastFrame(is_final_frame, cur_frm_idx)
            else:
                pass
        elif AudioChangeState.kChangeStateSpeech2Sil == state_change:
            self.continous_silence_frame_count = 0
            if (
                self.vad_state_machine
                == VadStateMachine.kVadInStateStartPointNotDetected
            ):
                pass
            elif self.vad_state_machine == VadStateMachine.kVadInStateInSpeechSegment:
                if (
                    cur_frm_idx - self.confirmed_start_frame + 1
                    > self.vad_opts.max_single_segment_time / frm_shift_in_ms
                ):
                    self.OnVoiceEnd(cur_frm_idx, False, False)
                    self.vad_state_machine = VadStateMachine.kVadInStateEndPointDetected
                elif not is_final_frame:
                    self.OnVoiceDetected(cur_frm_idx)
                else:
                    self.MaybeOnVoiceEndIfLastFrame(is_final_frame, cur_frm_idx)
            else:
                pass
        elif AudioChangeState.kChangeStateSpeech2Speech == state_change:
            self.continous_silence_frame_count = 0
            if self.vad_state_machine == VadStateMachine.kVadInStateInSpeechSegment:
                if (
                    cur_frm_idx - self.confirmed_start_frame + 1
                    > self.vad_opts.max_single_segment_time / frm_shift_in_ms
                ):
                    self.max_time_out = True
                    self.OnVoiceEnd(cur_frm_idx, False, False)
                    self.vad_state_machine = VadStateMachine.kVadInStateEndPointDetected
                elif not is_final_frame:
                    self.OnVoiceDetected(cur_frm_idx)
                else:
                    self.MaybeOnVoiceEndIfLastFrame(is_final_frame, cur_frm_idx)
            else:
                pass
        elif AudioChangeState.kChangeStateSil2Sil == state_change:
            self.continous_silence_frame_count += 1
            if (
                self.vad_state_machine
                == VadStateMachine.kVadInStateStartPointNotDetected
            ):
                # silence timeout, return zero length decision
                if (
                    (
                        self.vad_opts.detect_mode
                        == VadDetectMode.kVadSingleUtteranceDetectMode.value
                    )
                    and (
                        self.continous_silence_frame_count * frm_shift_in_ms
                        > self.vad_opts.max_start_silence_time
                    )
                ) or (is_final_frame and self.number_end_time_detected == 0):
                    for t in range(
                        self.lastest_confirmed_silence_frame + 1, cur_frm_idx
                    ):
                        self.OnSilenceDetected(t)
                    self.OnVoiceStart(0, True)
                    self.OnVoiceEnd(0, True, False)
                    self.vad_state_machine = VadStateMachine.kVadInStateEndPointDetected
                else:
                    if cur_frm_idx >= self.LatencyFrmNumAtStartPoint():
                        self.OnSilenceDetected(
                            cur_frm_idx - self.LatencyFrmNumAtStartPoint()
                        )
            elif self.vad_state_machine == VadStateMachine.kVadInStateInSpeechSegment:
                if (
                    self.continous_silence_frame_count * frm_shift_in_ms
                    >= self.max_end_sil_frame_cnt_thresh
                ):
                    lookback_frame = int(
                        self.max_end_sil_frame_cnt_thresh / frm_shift_in_ms
                    )
                    if self.vad_opts.do_extend:
                        lookback_frame -= int(
                            self.vad_opts.lookahead_time_end_point / frm_shift_in_ms
                        )
                        lookback_frame -= 1
                        lookback_frame = max(0, lookback_frame)
                    self.OnVoiceEnd(cur_frm_idx - lookback_frame, False, False)
                    self.vad_state_machine = VadStateMachine.kVadInStateEndPointDetected
                elif (
                    cur_frm_idx - self.confirmed_start_frame + 1
                    > self.vad_opts.max_single_segment_time / frm_shift_in_ms
                ):
                    self.OnVoiceEnd(cur_frm_idx, False, False)
                    self.vad_state_machine = VadStateMachine.kVadInStateEndPointDetected
                elif self.vad_opts.do_extend and not is_final_frame:
                    if self.continous_silence_frame_count <= int(
                        self.vad_opts.lookahead_time_end_point / frm_shift_in_ms
                    ):
                        self.OnVoiceDetected(cur_frm_idx)
                else:
                    self.MaybeOnVoiceEndIfLastFrame(is_final_frame, cur_frm_idx)
            else:
                pass

        if (
            self.vad_state_machine == VadStateMachine.kVadInStateEndPointDetected
            and self.vad_opts.detect_mode
            == VadDetectMode.kVadMutipleUtteranceDetectMode.value
        ):
            self.ResetDetection()

    @classmethod
    def from_model_dir(
        cls, model_dir: str = "checkpoints/iic/speech_fsmn_vad_zh-cn-16k-common-pytorch"
    ):
        config_path = Path(model_dir) / "config.yaml"
        config = read_yaml(config_path)
        return cls(config["model_conf"])

    @classmethod
    def from_config(cls, config: Dict[str, Any]):
        return cls(config)


def compute_audio_db(
    waveform: np.ndarray,
    sr: int = 16000,
    frame_length_ms: int = 25,
    frame_shift_ms: int = 10,
):
    """计算音频原始输入的分贝

    Args:
        waveform (np.ndarray): 音频波形
        sr (int, optional): 采样率. Defaults to 16000.
        frame_length_ms (int, optional): 帧长. Defaults to 25.
        frame_shift_ms (int, optional): 帧移. Defaults to 10.
    """
    frame_sample_length = int(frame_length_ms * sr / 1000)
    frame_shift_length = int(frame_shift_ms * sr / 1000)
    data = []
    for offset in range(
        0, waveform.shape[1] - frame_sample_length + 1, frame_shift_length
    ):
        d = waveform[0][offset : offset + frame_sample_length]
        data.append(d)
    data = np.stack(data)

    inputs = torch.from_numpy(data)
    decibel: torch.Tensor = compute_db(inputs)
    decibel = decibel.flatten().tolist()
    return decibel


@torch.compile
def compute_db(inputs: torch.Tensor) -> torch.Tensor:
    outputs = inputs.square().sum(dim=1, keepdim=True).add(1e-6).log10() * 10
    return outputs



from pathlib import Path
from typing import Any, Dict

import catalogue
import confection
from confection import Config


class registry(confection.registry):
    asr_models = catalogue.create("fasr", "asr_models", entry_points=True)
    vad_models = catalogue.create("fasr", "vad_models", entry_points=True)
    punc_models = catalogue.create("fasr", "punc_models", entry_points=True)
    lid_models = catalogue.create("fasr", "lid_models", entry_points=True)
    stream_vad_models = catalogue.create("fasr", "stream_vad_models", entry_points=True)
    stream_asr_models = catalogue.create("fasr", "stream_asr_models", entry_points=True)
    components = catalogue.create("fasr", "components", entry_points=True)
    stream_components = catalogue.create("fasr", "stream_components", entry_points=True)
    pipelines = catalogue.create("fasr", "pipelines", entry_points=True)
    pipes = catalogue.create("fasr", "pipes", entry_points=True)

    @classmethod
    def create(cls, registry_name: str, entry_points: bool = False) -> None:
        """Create a new custom registry."""
        if hasattr(cls, registry_name):
            raise ValueError(f"Registry '{registry_name}' already exists")
        reg = catalogue.create("fasr", registry_name, entry_points=entry_points)
        setattr(cls, registry_name, reg)

    @classmethod
    def resolve_from_dir(
        cls, save_dir: str, config: str = "config.cfg"
    ) -> Dict[str, Any]:
        """Resolve the registry from a directory."""
        save_dir = Path(save_dir)
        if not save_dir.exists():
            raise FileNotFoundError(f"Directory '{save_dir}' not found")
        config = save_dir / config
        if not config.exists():
            raise FileNotFoundError(f"Config file '{config}' not found")
        config = Config().from_disk(config)
        resolved = cls.resolve(config)
        return resolved


__all__ = ["Config", "registry"]



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

from dataclasses import dataclass
import logging
import os
import re
import time
import traceback

import torch

from .data.asr_feat import ASRFeatExtractor
from .models.fireredasr_aed import FireRedAsrAed
from .models.fireredasr_llm import FireRedAsrLlm
from .models.lstm_lm import LstmLm
from .models.param import count_model_parameters
from .tokenizer.aed_tokenizer import ChineseCharEnglishSpmTokenizer
from .tokenizer.llm_tokenizer import LlmTokenizerWrapper

logger = logging.getLogger(__name__)


@dataclass
class FireRedAsr2Config:
    use_gpu: bool = True
    use_half: bool = False
    use_bfloat16: bool = False
    beam_size: int = 3
    nbest: int = 1
    decode_max_len: int = 0
    softmax_smoothing: float = 1.25
    aed_length_penalty: float = 0.6
    eos_penalty: float = 1.0
    return_timestamp: bool = False
    decode_min_len: bool = 0
    repetition_penalty: float = 1.0
    llm_length_penalty: float = 0.0
    temperature: float = 1.0
    elm_dir: str = ""
    elm_weight: float = 0.0

    def __post_init__(self):
        pass


class FireRedAsr2:
    @classmethod
    def from_pretrained(cls, asr_type, model_dir, config=FireRedAsr2Config()):
        assert asr_type in ["aed", "llm"]

        cmvn_path = os.path.join(model_dir, "cmvn.ark")
        feat_extractor = ASRFeatExtractor(cmvn_path)

        if asr_type == "aed":
            model_path = os.path.join(model_dir, "model.pth.tar")
            dict_path = os.path.join(model_dir, "dict.txt")
            spm_model = os.path.join(model_dir, "train_bpe1000.model")
            model = load_fireredasr_aed_model(model_path)
            tokenizer = ChineseCharEnglishSpmTokenizer(dict_path, spm_model)
        elif asr_type == "llm":
            model_path = os.path.join(model_dir, "model.pth.tar")
            encoder_path = os.path.join(model_dir, "asr_encoder.pth.tar")
            llm_dir = os.path.join(model_dir, "Qwen2-7B-Instruct")
            model, tokenizer = load_firered_llm_model_and_tokenizer(
                model_path, encoder_path, llm_dir
            )
        elm = None
        if config.elm_dir:
            assert os.path.exists(config.elm_dir), f"{config.elm_dir}"
            model_path = os.path.join(config.elm_dir, "model.pth.tar")
            elm = load_lstm_lm(model_path)
            elm.eval()
            logger.info(elm)
        count_model_parameters(model)
        model.eval()
        return cls(asr_type, feat_extractor, model, tokenizer, elm, config)

    def __init__(self, asr_type, feat_extractor, model, tokenizer, elm, config):
        self.asr_type = asr_type
        self.feat_extractor = feat_extractor
        self.model = model
        self.tokenizer = tokenizer
        self.elm = elm
        self.config = config
        logger.info(self.config)
        if self.config.use_gpu:
            if self.config.use_bfloat16:
                self.model.to(torch.bfloat16)
            elif self.config.use_half:
                self.model.half()
            self.model.cuda()
            if self.elm:
                self.elm.cuda()
        else:
            self.model.cpu()

    @torch.no_grad()
    def transcribe(self, batch_uttid, batch_wav_path):
        batch_uttid_origin = batch_uttid
        try:
            feats, lengths, durs, batch_wav_path, batch_uttid = self.feat_extractor(
                batch_wav_path, batch_uttid
            )
            if feats is None:
                return [{"uttid": uttid, "text": ""} for uttid in batch_uttid_origin]
        except:
            traceback.print_exc()
            return [{"uttid": uttid, "text": ""} for uttid in batch_uttid_origin]
        total_dur = sum(durs)
        if self.config.use_gpu:
            feats, lengths = feats.cuda(), lengths.cuda()
            model_dtype = next(self.model.parameters()).dtype
            if feats.dtype != model_dtype:
                feats = feats.to(model_dtype)

        if self.asr_type == "aed":
            start_time = time.time()

            try:
                hyps = self.model.transcribe(
                    feats,
                    lengths,
                    self.config.beam_size,
                    self.config.nbest,
                    self.config.decode_max_len,
                    self.config.softmax_smoothing,
                    self.config.aed_length_penalty,
                    self.config.eos_penalty,
                    self.config.return_timestamp,
                    self.elm,
                    self.config.elm_weight,
                )
            except Exception as e:
                traceback.print_exc()
                hyps = []

            elapsed = time.time() - start_time
            rtf = elapsed / total_dur if total_dur > 0 else 0

            results = []
            for uttid, wav, hyp, dur in zip(batch_uttid, batch_wav_path, hyps, durs):
                hyp = hyp[0]  # only return 1-best
                hyp_ids = [int(id) for id in hyp["yseq"].cpu()]
                text = self.tokenizer.detokenize(hyp_ids)
                text = re.sub(r"(<blank>)|(<sil>)", "", text)
                results.append(
                    {
                        "uttid": uttid,
                        "text": text.lower(),
                        "confidence": round(hyp["confidence"].cpu().item(), 3),
                        "dur_s": round(dur, 3),
                        "rtf": f"{rtf:.4f}",
                    }
                )
                if type(wav) == str:
                    results[-1]["wav"] = wav
                if self.config.return_timestamp:
                    results[-1]["timestamp"] = self._get_and_fix_timestamp(
                        hyp, hyp_ids, dur
                    )
            return results

        elif self.asr_type == "llm":
            input_ids, attention_mask, _, _ = LlmTokenizerWrapper.preprocess_texts(
                origin_texts=[""] * feats.size(0),
                tokenizer=self.tokenizer,
                max_len=128,
                decode=True,
            )
            if self.config.use_gpu:
                input_ids = input_ids.cuda()
                attention_mask = attention_mask.cuda()
            start_time = time.time()

            try:
                generated_ids = self.model.transcribe(
                    feats,
                    lengths,
                    input_ids,
                    attention_mask,
                    self.config.beam_size,
                    self.config.decode_max_len,
                    self.config.decode_min_len,
                    self.config.repetition_penalty,
                    self.config.llm_length_penalty,
                    self.config.temperature,
                )
                texts = self.tokenizer.batch_decode(
                    generated_ids, skip_special_tokens=True
                )
            except Exception as e:
                logger.error(f"LLM transcribe error: {e}")
                texts = []

            elapsed = time.time() - start_time
            rtf = elapsed / total_dur if total_dur > 0 else 0
            results = []
            for uttid, wav, text in zip(batch_uttid, batch_wav_path, texts):
                results.append(
                    {"uttid": uttid, "text": text.lower(), "rtf": f"{rtf:.4f}"}
                )
                if type(wav) == str:
                    results[-1]["wav"] = wav
            return results

    def _get_and_fix_timestamp(self, hyp, hyp_ids, dur):
        r3 = lambda x: round(x, 3)
        if "timestamp" not in hyp or hyp["timestamp"] is None:
            timestamp = []
            avg_dur = dur / len(hyp_ids) if len(hyp_ids) > 0 else 0
            last_end = dur
            for i, hyp_id in enumerate(hyp_ids):
                token = self.tokenizer.detokenize([hyp_id], "", False)
                start = min(max(0, i * avg_dur), last_end)
                end = min((i + 1) * avg_dur, dur)
                last_end = end
                timestamp.append([token.lower(), r3(start), r3(end)])
        else:
            starts, ends = hyp["timestamp"]
            timestamp = []
            last_end = dur
            SHIFT = 0.06  # shift 40ms
            for hyp_id, start, end in zip(hyp_ids, starts, ends):
                token = self.tokenizer.detokenize([hyp_id], "", False)
                start = min(max(0, start - SHIFT), last_end)
                end = min(max(0, end - SHIFT), dur)
                last_end = end
                timestamp.append([token.lower(), r3(start), r3(end)])
        # Fix case: start == dur and end == dur
        for i in range(len(timestamp)):
            idx = -(i + 1)
            _, start, end = timestamp[idx]
            if abs(dur - start) < 0.001:
                logger.info(f"start before {timestamp[idx]}")
                timestamp[idx][1] = dur - (i + 1) * 0.001
                logger.info(f"start after {timestamp[idx]}")
            if i != 0 and abs(dur - end) < 0.001:
                logger.info(f"end before {timestamp[idx]}")
                timestamp[idx][2] = dur - i * 0.001
                logger.info(f"end after {timestamp[idx]}")
        timestamp = self.tokenizer.merge_spm_timestamp(timestamp)
        return timestamp


def load_fireredasr_aed_model(model_path):
    package = torch.load(
        model_path, map_location=lambda storage, loc: storage, weights_only=False
    )
    model = FireRedAsrAed.from_args(package["args"])
    model.load_state_dict(package["model_state_dict"], strict=False)
    return model


def _merge_lora_state_dict(model, state_dict, lora_alpha=16, lora_r=64):
    """Merge LoRA A/B weights into the base model without requiring peft."""
    scaling = lora_alpha / lora_r
    param_dict = dict(model.named_parameters())
    merged = 0
    for key in list(state_dict.keys()):
        if ".lora_A." not in key:
            continue
        b_key = key.replace(".lora_A.", ".lora_B.")
        if b_key not in state_dict:
            continue
        base_key = key.replace(".lora_A.default.weight", ".weight")
        base_key = base_key.replace("base_model.model.", "", 1)
        if base_key not in param_dict:
            logger.warning("LoRA target not found: %s", base_key)
            continue
        lora_a = state_dict[key]
        lora_b = state_dict[b_key]
        param = param_dict[base_key]
        param.data.add_((lora_b @ lora_a).to(param.dtype) * scaling)
        merged += 1
    logger.info("Merged %d LoRA adapter pairs into base model", merged)


def load_firered_llm_model_and_tokenizer(model_path, encoder_path, llm_dir):
    package = torch.load(
        model_path, map_location=lambda storage, loc: storage, weights_only=False
    )
    package["args"].encoder_path = encoder_path
    package["args"].llm_dir = llm_dir
    package["args"].use_fp16 = True
    model = FireRedAsrLlm.from_args(package["args"])
    model.load_state_dict(package["model_state_dict"], strict=False)
    _merge_lora_state_dict(model, package["model_state_dict"])
    tokenizer = LlmTokenizerWrapper.build_llm_tokenizer(llm_dir)
    return model, tokenizer


def load_lstm_lm(model_path):
    package = torch.load(
        model_path, map_location=lambda storage, loc: storage, weights_only=False
    )
    model = LstmLm.from_args(package["args"])
    model.load_state_dict(package["model_state_dict"], strict=False)
    return model



from abc import abstractmethod
from pathlib import Path
from typing import Dict, Iterable, List

from docarray.typing import AnyTensor
from pydantic import ConfigDict
from typing_extensions import Self

from .config import Config
from .data import AudioChunk, AudioSpanList, AudioToken, AudioTokenList, Waveform
from .utils.base import CheckpointMixin, IOMixin


class Model(CheckpointMixin, IOMixin):
    model_config: ConfigDict = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True
    )

    def get_config(self) -> Config:
        raise NotImplementedError

    def load(self, save_dir: str | Path, **kwargs) -> Self:
        raise NotImplementedError

    def save(self, save_dir: str | Path, **kwargs) -> None:
        raise NotImplementedError

    def from_checkpoint(self, checkpoint_dir: str | Path, **kwargs) -> Self:
        raise NotImplementedError


class StreamModel(Model):
    model_config: ConfigDict = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True
    )

    states: Dict = {}

    def reset(self):
        self.states.clear()

    def remove_state(self, key: str):
        self.states.pop(key, None)

    def get_state(self, key: str):
        return self.states.get(key, None)


class ASRModel(Model):
    """语音识别模型基类"""

    @abstractmethod
    def transcribe(
        self,
        batch: List[AnyTensor],
        **kwargs,
    ) -> List[AudioTokenList]:
        raise NotImplementedError


class StreamASRModel(StreamModel):
    """流式语音识别模型基类"""

    chunk_size_ms: int = None

    @abstractmethod
    def transcribe_chunk(self, chunk: AudioChunk) -> Iterable[AudioToken]:
        raise NotImplementedError


class PuncModel(Model):
    @abstractmethod
    def restore(self, text: str) -> AudioSpanList:
        raise NotImplementedError


class LIDModel(Model):
    @abstractmethod
    def identify(self, waveform: Waveform) -> str:
        """识别音频的语言，返回语言标识字符串（如 'zh', 'en'）。"""
        raise NotImplementedError


class VADModel(Model):
    @abstractmethod
    def detect(self, waveform: Waveform) -> AudioSpanList:
        raise NotImplementedError


class StreamVADModel(StreamModel):
    chunk_size_ms: int = 100
    sample_rate: int = 16000
    max_end_silence_time: int = 500
    db_threshold: int | None = None

    @abstractmethod
    def detect_chunk(chunk: AudioChunk) -> Iterable[AudioChunk]:
        raise NotImplementedError

    @property
    def chunk_size(self) -> int:
        return self.chunk_size_ms * self.sample_rate // 1000



from fasr.config import Config, registry


def load(save_dir: str, type: str = "component"):
    from pathlib import Path

    save_dir = Path(save_dir)
    if not save_dir.exists():
        raise FileNotFoundError(f"Directory {save_dir} does not exist")
    config_path = save_dir / "config.cfg"
    config = Config().from_disk(config_path)
    obj = registry.resolve(config)[type]
    loaded_obj = obj.load(save_dir)
    return loaded_obj


def load_plugin(group: str, name: str | None = None) -> None:
    from importlib.metadata import entry_points

    for plugin in entry_points(group=group):
        if name is None:
            plugin.load()
        else:
            if plugin.name == name:
                plugin.load()
                return
    raise ValueError(f"Plugin {name} not found in group {group}")



from pathlib import Path
from typing import List

from funasr import AutoModel
import numpy as np
import torch
from typing_extensions import Self

from fasr.config import registry
from fasr.data import AudioToken, AudioTokenList
from fasr.model import ASRModel


@registry.asr_models.register("paraformer")
class ParaformerForASR(ASRModel):
    checkpoint: str = "iic/speech_paraformer-large-vad-punc_asr_nat-zh-cn-16k-common-vocab8404-pytorch"
    endpoint: str = "modelscope"

    paraformer: AutoModel | None = None

    def transcribe(
        self, batch: List[np.ndarray | torch.Tensor], **kwargs
    ) -> List[List[AudioToken]]:
        fs = kwargs.get("sample_rate", 16000)
        results = self.paraformer.generate(input=batch, fs=fs)
        batch_tokens = []
        for result in results:
            tokens = AudioTokenList()
            result_text = result["text"]
            if result_text:
                texts = result["text"].split(" ")
                timestamps = result["timestamp"]
                assert len(texts) == len(timestamps), (
                    f"Texts and timestamps have different lengths: {len(texts)} != {len(timestamps)}"
                )
                for token_text, timestamp in zip(texts, timestamps):
                    start_ms = timestamp[0]
                    end_ms = timestamp[1]
                    token = AudioToken(
                        start_ms=start_ms, end_ms=end_ms, text=token_text
                    )
                    assert token.end_ms - token.start_ms > 0, f"{token}"
                    tokens.append(token)
            batch_tokens.append(tokens)
        return batch_tokens

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        disable_update: bool = True,
        disable_log: bool = True,
        disable_pbar: bool = True,
        **kwargs,
    ) -> Self:
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        self.paraformer = AutoModel(
            model=checkpoint_dir,
            disable_update=disable_update,
            disable_log=disable_log,
            disable_pbar=disable_pbar,
            **kwargs,
        )
        self.paraformer.kwargs["batch_size"] = 10000  # enable batch processing
        return self

    def get_config(self):
        raise NotImplementedError

    def load(self, save_dir, **kwargs):
        raise NotImplementedError

    def save(self, save_dir, **kwargs):
        raise NotImplementedError


@registry.asr_models.register("seaco_paraformer")
class SeacoParaformerForASR(ASRModel):
    checkpoint: str = (
        "iic/speech_seaco_paraformer_large_asr_nat-zh-cn-16k-common-vocab8404-pytorch"
    )
    endpoint: str = "modelscope"

    seaco_paraformer: AutoModel | None = None

    def transcribe(
        self, batch: List[np.ndarray | torch.Tensor], **kwargs
    ) -> List[List[AudioToken]]:
        fs = kwargs.get("sample_rate", 16000)
        hotwords = kwargs.get("hotwords", None)
        hotword = " ".join(hotwords) if hotwords else None
        results = self.seaco_paraformer.generate(input=batch, fs=fs, hotword=hotword)
        batch_tokens = []
        for result in results:
            tokens = AudioTokenList()
            result_text = result["text"]
            if result_text:
                texts = result["text"].split(" ")
            else:
                texts = []
            timestamps = result["timestamp"]
            assert len(texts) == len(timestamps), (
                f"Texts and timestamps have different lengths: {len(texts)} != {len(timestamps)}"
            )
            for token_text, timestamp in zip(texts, timestamps):
                start_ms = timestamp[0]
                end_ms = timestamp[1]
                token = AudioToken(start_ms=start_ms, end_ms=end_ms, text=token_text)
                assert token.end_ms - token.start_ms > 0, f"{token}"
                tokens.append(token)
            batch_tokens.append(tokens)
        return batch_tokens

    def from_checkpoint(
        self,
        checkpoint_dir: str | Path | None = None,
        disable_update: bool = True,
        disable_log: bool = True,
        disable_pbar: bool = True,
        **kwargs,
    ) -> Self:
        if not checkpoint_dir:
            checkpoint_dir = self.download_checkpoint()
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_dir}")
        model = AutoModel(
            model=checkpoint_dir,
            disable_update=disable_update,
            disable_log=disable_log,
            disable_pbar=disable_pbar,
            **kwargs,
        )
        self.seaco_paraformer = model
        self.seaco_paraformer.model.generate_hotwords_list = self.generate_hotwords
        self.seaco_paraformer.kwargs["batch_size"] = 10000  # enable batch processing
        return self

    def generate_hotwords(
        self, hotword_list_or_file: str | None = None, **kwargs
    ) -> List[List[int]] | None:
        """Generate hotwords list.

        Args:
            hotword_list_or_file (str): Hotword string. e.g. "你好 你好吗"
        """
        if not hotword_list_or_file:
            return None
        hotwords = hotword_list_or_file.strip().split(" ")
        if len(hotwords) == 0:
            return None
        all_ids = []
        for hotword in hotwords:
            tokens = [
                s for s in hotword
            ]  # todo: better tokenization. now it's just char-level.
            all_ids.append(self.tokens2ids(tokens))
        all_ids.append(self.tokens2ids(["<s>"]))
        return all_ids

    def tokens2ids(self, tokens: List[AudioToken], **kwargs) -> List[int]:
        return self.tokenizer.tokens2ids(tokens)

    @property
    def tokenizer(self):
        return self.seaco_paraformer.kwargs["tokenizer"]

    def get_config(self):
        raise NotImplementedError

    def load(self, save_dir, **kwargs):
        raise NotImplementedError

    def save(self, save_dir, **kwargs):
        raise NotImplementedError



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

import torch

from .module.conformer_encoder import ConformerEncoder
from .module.transformer_decoder import TransformerDecoder


class FireRedLidAed(torch.nn.Module):
    @classmethod
    def from_args(cls, args):
        return cls(args)

    def __init__(self, args):
        super().__init__()
        self.sos_id = args.sos_id
        self.eos_id = args.eos_id

        self.encoder = ConformerEncoder(
            args.idim,
            args.n_layers_enc,
            args.n_head,
            args.d_model,
            args.residual_dropout,
            args.dropout_rate,
            args.kernel_size,
            args.pe_maxlen,
        )

        self.lid_decoder = TransformerDecoder(
            args.sos_id,
            args.eos_id,
            args.pad_id,
            args.lid_odim,
            args.n_layers_lid_dec,
            args.n_head,
            args.d_model,
            args.residual_dropout,
            args.pe_maxlen,
        )

    def process(
        self,
        padded_input,
        input_lengths,
        beam_size=3,
        nbest=1,
        decode_max_len=2,
        softmax_smoothing=1.25,
        length_penalty=0.6,
        eos_penalty=1.0,
    ):
        enc_outputs, enc_lengths, enc_mask = self.encoder(padded_input, input_lengths)
        nbest_hyps = self.lid_decoder.batch_beam_search(
            enc_outputs,
            enc_mask,
            beam_size,
            nbest,
            decode_max_len,
            softmax_smoothing,
            length_penalty,
            eos_penalty,
        )
        return nbest_hyps



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Yan Jia)

import math
import os

import kaldi_native_fbank as knf
import kaldiio
import numpy as np
import torch


class FeatExtractor:
    def __init__(self, kaldi_cmvn_file):
        self.cmvn = CMVN(kaldi_cmvn_file) if kaldi_cmvn_file != "" else None
        self.fbank = KaldifeatFbank(
            num_mel_bins=80, frame_length=25, frame_shift=10, dither=0.0
        )

    def __call__(self, wav_paths, wav_uttids):
        feats = []
        durs = []
        return_wav_paths = []
        return_wav_uttids = []

        wav_datas = []
        if isinstance(wav_paths[0], str):
            for wav_path in wav_paths:
                sample_rate, wav_np = kaldiio.load_mat(wav_path)
                wav_datas.append([sample_rate, wav_np])
        else:
            wav_datas = wav_paths

        for (sample_rate, wav_np), path, uttid in zip(wav_datas, wav_paths, wav_uttids):
            dur = wav_np.shape[0] / sample_rate
            fbank = self.fbank((sample_rate, wav_np))
            if fbank.shape[0] < 1:
                continue
            if self.cmvn is not None:
                fbank = self.cmvn(fbank)
            fbank = torch.from_numpy(fbank).float()
            feats.append(fbank)
            durs.append(dur)
            return_wav_paths.append(path)
            return_wav_uttids.append(uttid)
        if len(feats) > 0:
            lengths = torch.tensor([feat.size(0) for feat in feats]).long()
            feats_pad = self.pad_feat(feats, 0.0)
        else:
            lengths, feats_pad = None, None
        return feats_pad, lengths, durs, return_wav_paths, return_wav_uttids

    def pad_feat(self, xs, pad_value):
        n_batch = len(xs)
        max_len = max([xs[i].size(0) for i in range(n_batch)])
        pad = (
            torch.ones(n_batch, max_len, *xs[0].size()[1:])
            .to(xs[0].device)
            .to(xs[0].dtype)
            .fill_(pad_value)
        )
        for i in range(n_batch):
            pad[i, : xs[i].size(0)] = xs[i]
        return pad


class CMVN:
    def __init__(self, kaldi_cmvn_file):
        self.dim, self.means, self.inverse_std_variences = self.read_kaldi_cmvn(
            kaldi_cmvn_file
        )

    def __call__(self, x, is_train=False):
        assert x.shape[-1] == self.dim, "CMVN dim mismatch"
        out = x - self.means
        out = out * self.inverse_std_variences
        return out

    def read_kaldi_cmvn(self, kaldi_cmvn_file):
        assert os.path.exists(kaldi_cmvn_file)
        stats = kaldiio.load_mat(kaldi_cmvn_file)
        assert stats.shape[0] == 2
        dim = stats.shape[-1] - 1
        count = stats[0, dim]
        assert count >= 1
        floor = 1e-20
        means = []
        inverse_std_variences = []
        for d in range(dim):
            mean = stats[0, d] / count
            means.append(mean.item())
            varience = (stats[1, d] / count) - mean * mean
            if varience < floor:
                varience = floor
            istd = 1.0 / math.sqrt(varience)
            inverse_std_variences.append(istd)
        return dim, np.array(means), np.array(inverse_std_variences)


class KaldifeatFbank:
    def __init__(self, num_mel_bins=80, frame_length=25, frame_shift=10, dither=1.0):
        self.dither = dither
        opts = knf.FbankOptions()
        opts.frame_opts.dither = dither
        opts.mel_opts.num_bins = num_mel_bins
        opts.frame_opts.snip_edges = True
        opts.mel_opts.debug_mel = False
        self.opts = opts

    def __call__(self, wav, is_train=False):
        if type(wav) is str:
            sample_rate, wav_np = kaldiio.load_mat(wav)
        elif type(wav) in [tuple, list] and len(wav) == 2:
            sample_rate, wav_np = wav
        assert len(wav_np.shape) == 1

        dither = self.dither if is_train else 0.0
        self.opts.frame_opts.dither = dither
        fbank = knf.OnlineFbank(self.opts)

        fbank.accept_waveform(sample_rate, wav_np.tolist())
        feat = []
        for i in range(fbank.num_frames_ready):
            feat.append(fbank.get_frame(i))
        if len(feat) == 0:
            print("Check data, len(feat) == 0", wav, flush=True)
            return np.zeros((0, self.opts.mel_opts.num_bins))
        feat = np.vstack(feat)
        return feat



from __future__ import annotations

from typing import List

from docarray import BaseDoc
from docarray.typing import ID, AnyTensor
import librosa
import numpy as np
from pydantic import ConfigDict, Field
from typing_extensions import Self


class Waveform(BaseDoc):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="ignore")

    id: ID | None = None
    data: np.ndarray | None = Field(None, description="Waveform data.")
    sample_rate: int | None = Field(None, description="Sample rate of the waveform.")
    is_normalized: bool | None = Field(
        None,
        description=(
            "Whether the waveform data is normalized to [-1.0, 1.0] float range. "
            "True: data is float32 in [-1.0, 1.0], typically from librosa.load(). "
            "False/None: data is raw PCM integers (e.g. int16 in [-32768, 32767])."
        ),
    )
    feats: AnyTensor | None = Field(
        None, description="Pre-extracted features of the waveform (e.g. fbank)."
    )
    scores: AnyTensor | None = Field(
        None, description="Model output scores associated with the waveform."
    )

    def to_int16_pcm(self) -> np.ndarray:
        """Convert waveform data to int16 PCM format.

        If is_normalized is True, scales float [-1.0, 1.0] to int16 [-32768, 32767].
        If is_normalized is None, falls back to dtype check: float types are treated
        as normalized.
        Otherwise returns data as-is (already PCM).
        """
        wav = self.data
        if wav.ndim > 1:
            wav = wav.flatten()
        need_convert = self.is_normalized
        if need_convert is None:
            need_convert = np.issubdtype(wav.dtype, np.floating)
        if need_convert:
            wav = (wav * 32768).astype(np.int16)
        return wav

    def resample(self, sample_rate: int) -> Self:
        """Resample the waveform to the target sample rate.

        Args:
            sample_rate (int): Target sample rate.
        """
        self.data = librosa.resample(
            self.data, orig_sr=self.sample_rate, target_sr=sample_rate
        )
        self.sample_rate = sample_rate
        return self

    @classmethod
    def concatenate(cls, waveforms: List[Waveform]) -> Self:
        """Concatenate the waveform with another waveform."""
        data = np.concatenate([waveform.data for waveform in waveforms])
        sample_rate = waveforms[0].sample_rate
        return cls(data=data, sample_rate=sample_rate)

    def append(self, waveform: Waveform) -> Self:
        """Append the waveform with another waveform."""
        assert self.sample_rate == waveform.sample_rate, (
            "Only support the same sample rate."
        )
        data = np.concatenate([self.data, waveform.data])
        self.data = data
        return self

    def flatten(self) -> Self:
        """Flatten the waveform."""
        self.data = self.data.flatten()
        return self

    def as_1d(self, clear: bool = True) -> Self:
        """Flatten the waveform."""
        self.data = self.data.flatten()
        if clear:
            self.feats = None
            self.scores = None
        return self

    def display(self) -> Self:
        """Display the waveform."""
        from IPython.display import Audio

        return Audio(data=self.data, rate=self.sample_rate)

    def select_by_ms(self, start: float | None, end: float | None) -> Waveform:
        """Select the waveform by milliseconds."""
        if start is None:
            start = 0
        if end is None:
            end = self.duration_ms
        start_idx = int(start * self.sample_rate / 1000)
        end_idx = int(end * self.sample_rate / 1000)
        return self[start_idx:end_idx]

    @property
    def ndim(self):
        """Get the number of dimensions of the waveform."""
        return self.data.ndim

    @property
    def duration(self):
        """Get the duration of the waveform in seconds."""
        return librosa.get_duration(y=self.data, sr=self.sample_rate)

    @property
    def duration_s(self):
        """Get the duration of the waveform in seconds."""
        return round(librosa.get_duration(y=self.data, sr=self.sample_rate), 3)

    @property
    def duration_ms(self):
        """Get the duration of the waveform in milliseconds."""
        return self.duration * 1000

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index: int | slice) -> Waveform:
        """Get the value at the index. e.g. waveform[0], waveform[1:10]"""
        data = self.data[index]
        return Waveform(data=data, sample_rate=self.sample_rate)



import re
from typing import Dict, List, Union

import editdistance
import numpy as np

from fasr.normalizer import Normalizer


def compute_char_error_rate(
    x: str,
    y: str,
    ignore_space: bool = True,
    ignore_punc: bool = True,
    normalize: bool = False,
) -> float:
    """Compute the character error rate between two strings.

    Args:
        x (str): the source string
        y (str): the target string
        ignore_space (bool, optional): whether to ignore spaces. Defaults to True.
        ignore_punc (bool, optional): whether to ignore punctuation. Defaults to True.
        normalizer (Normalizer, optional): the normalizer used to normalize the strings, must have a normalize method. Defaults to None.

    Returns:
        float: the character error rate
    """
    assert isinstance(x, str) and isinstance(y, str)
    normalizer = (
        Normalizer(remove_erhua=True, enable_0_to_9=True, lang="zh", operator="itn")
        if normalize
        else None
    )
    if normalizer is not None:
        x = normalizer.normalize(x)
        y = normalizer.normalize(y)
    x = x.replace(" ", "") if ignore_space else x
    x = re.sub(r"[^\w\s]", "", x) if ignore_punc else x
    y = y.replace(" ", "") if ignore_space else y
    y = re.sub(r"[^\w\s]", "", y) if ignore_punc else y
    if len(y) == 0:
        if len(x) == 0:
            return 1.0
        else:
            return 0.0
    return round(editdistance.eval(x, y) / len(y), 4)


def match_token(hyp: str, ref: str) -> List:
    """Match the tokens between two strings.

    Args:
        hyp (str): the hypothesis string
        ref (str): the reference string

    Returns:
        List: the list of matched results, eg. [(0, 0), (1, 1), (2, 2), (3, 3)], where the first element is the index of the reference string and the second element is the index of the hypothesis string.
    """
    hyp = list(map(lambda x: x.lower(), hyp))
    ref = list(map(lambda x: x.lower(), ref))

    len_hyp = len(hyp)
    len_ref = len(ref)

    cost_matrix = np.zeros((len_hyp + 1, len_ref + 1), dtype=np.int16)
    ops_matrix = np.zeros((len_hyp + 1, len_ref + 1), dtype=np.int8)

    for i in range(len_hyp + 1):
        cost_matrix[i][0] = i
    for j in range(len_ref + 1):
        cost_matrix[0][j] = j

    for i in range(1, len_hyp + 1):
        for j in range(1, len_ref + 1):
            if hyp[i - 1] == ref[j - 1]:
                cost_matrix[i][j] = cost_matrix[i - 1][j - 1]
            else:
                substitution = cost_matrix[i - 1][j - 1] + 1
                insertion = cost_matrix[i - 1][j] + 1
                deletion = cost_matrix[i][j - 1] + 1

                compare_val = [substitution, insertion, deletion]

                min_val = min(compare_val)
                operation_idx = compare_val.index(min_val) + 1
                cost_matrix[i][j] = min_val
                ops_matrix[i][j] = operation_idx

    match_idx = []
    i = len_hyp
    j = len_ref
    rst = {"nwords": len_ref, "cor": 0, "wrong": 0, "ins": 0, "del": 0, "sub": 0}
    while i >= 0 or j >= 0:
        i_idx = max(0, i)
        j_idx = max(0, j)

        if ops_matrix[i_idx][j_idx] == 0:  # correct
            if i - 1 >= 0 and j - 1 >= 0:
                match_idx.append((j - 1, i - 1))
                rst["cor"] += 1

            i -= 1
            j -= 1

        elif ops_matrix[i_idx][j_idx] == 2:  # insert
            i -= 1
            rst["ins"] += 1

        elif ops_matrix[i_idx][j_idx] == 3:  # delete
            j -= 1
            rst["del"] += 1

        elif ops_matrix[i_idx][j_idx] == 1:  # substitute
            i -= 1
            j -= 1
            rst["sub"] += 1

        if i < 0 and j >= 0:
            rst["del"] += 1
        elif j < 0 and i >= 0:
            rst["ins"] += 1

    match_idx.reverse()
    wrong_cnt = cost_matrix[len_hyp][len_ref]
    rst["wrong"] = wrong_cnt
    return match_idx


def streaming_delay(
    t_send: List[Dict[str, float]],
    t_response: List[Dict[str, float]],
    return_results: bool = False,
) -> List[Dict[str, Union[str, float]]] | float:
    """calculate the delay between the fire and response tokens

    Args:
        t_send (Dict[str, float]): send tokens with timestamp. eg. [{'token': 'a', 'timestamp': 0.1}, {'token': 'b', 'timestamp': 0.2}]
        t_response (Dict[str, float]): response tokens with timestamp. eg. [{'token': 'a', 'timestamp': 0.3}, {'token': 'b', 'timestamp': 0.4}]
        return_results (bool, optional): whether to return the results. Defaults to False.

    Returns:
        List[float]: the delay between the fire and response tokens. eg. [{'token': 'a', 'send': 0.1, 'response': 0.3, 'delay': 0.2}, {'token': 'b', 'send': 0.2, 'response': 0.4, 'delay': 0.2}]
    """
    ref = "".join([t["token"] for t in t_send])
    hyp = "".join([t["token"] for t in t_response])

    matched = match_token(hyp, ref)
    delay = []
    for m in matched:
        t = {}
        t["token"] = t_send[m[0]]["token"]
        t["send"] = t_send[m[0]]["timestamp"]
        t["response"] = t_response[m[1]]["timestamp"]
        t["delay"] = t["response"] - t["send"]
        delay.append(t)
    if return_results:
        return delay
    else:
        return round(sum([t["delay"] for t in delay]) / len(delay), 4)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Literal

from pydantic import BaseModel

from fasr.config import Config

FASR_DEFAULT_CACHE_DIR = Path.home() / ".cache" / "fasr" / "checkpoints"

_cache_dir: Path = FASR_DEFAULT_CACHE_DIR


def set_cache_dir(path: str | Path) -> None:
    global _cache_dir
    _cache_dir = Path(path)


def get_cache_dir() -> Path:
    return _cache_dir


def download(
    repo_id: str,
    revision: str = None,
    cache_dir: str | Path | None = None,
    endpoint: Literal["modelscope", "huggingface", "hf-mirror"] = "modelscope",
) -> None:
    """Download model from modelscope"""
    cache_dir = Path(cache_dir) if cache_dir is not None else get_cache_dir()
    if endpoint == "hf-mirror":
        from huggingface_hub import snapshot_download

        _ = snapshot_download(
            repo_id=repo_id,
            local_dir=cache_dir / repo_id,
            revision=revision,
            local_dir_use_symlinks=False,
            endpoint="https://hf-mirror.com",
        )
    if endpoint == "huggingface":
        from huggingface_hub import snapshot_download

        _ = snapshot_download(
            repo_id=repo_id,
            local_dir=cache_dir / repo_id,
            revision=revision,
            local_dir_use_symlinks=False,
        )

    elif endpoint == "modelscope":
        from modelscope import snapshot_download

        _ = snapshot_download(model_id=repo_id, cache_dir=cache_dir, revision=revision)


class IOMixin(BaseModel, ABC):
    @abstractmethod
    def save(self, save_dir: str):
        raise NotImplementedError

    @abstractmethod
    def load(self, save_dir: str):
        raise NotImplementedError

    @abstractmethod
    def get_config(self) -> Config:
        raise NotImplementedError


class CheckpointMixin(BaseModel, ABC):
    cache_dir: str | Path | None = None
    checkpoint: str | None = None
    endpoint: Literal["modelscope", "huggingface", "hf-mirror"] = "hf-mirror"

    @property
    def resolved_cache_dir(self) -> Path:
        if self.cache_dir is not None:
            return Path(self.cache_dir)
        return get_cache_dir()

    @abstractmethod
    def from_checkpoint(self, checkpoint_dir: str, device: str, **kwargs):
        raise NotImplementedError

    def download_checkpoint(self, revision: str = None) -> Path:
        if self.checkpoint is None:
            raise ValueError("checkpoint is None")
        checkpoint_dir = self.resolved_cache_dir / self.checkpoint
        if not checkpoint_dir.exists():
            download(
                repo_id=self.checkpoint,
                revision=revision,
                cache_dir=self.resolved_cache_dir,
                endpoint=self.endpoint,
            )
        return checkpoint_dir

    @property
    def checkpoint_dir(self) -> Path | None:
        if self.checkpoint is None:
            return None
        return self.resolved_cache_dir / self.checkpoint

    @property
    def default_checkpoint_dir(self) -> Path:
        return self.resolved_cache_dir / self.checkpoint


def clear_cache(cache_dir: str | Path | None = None):
    """清空缓存目录

    Args:
        cache_dir (str | Path, optional): 缓存目录. 默认使用 get_cache_dir() 返回的全局缓存目录.
    """
    cache_dir = Path(cache_dir) if cache_dir is not None else get_cache_dir()
    if cache_dir.exists():
        for item in cache_dir.iterdir():
            if item.is_dir():
                for sub_item in item.iterdir():
                    sub_item.unlink()
                item.rmdir()
            else:
                item.unlink()
    else:
        cache_dir.mkdir(parents=True)
    return cache_dir



from pydantic import ConfigDict

from fasr.utils.base import IOMixin


class BaseRuntime(IOMixin):
    model_config: ConfigDict = ConfigDict(arbitrary_types_allowed=True)

    def run(self, **kwargs):
        raise NotImplementedError

    def save(self, save_dir):
        raise NotImplementedError

    def load(self, save_dir):
        raise NotImplementedError



from abc import ABC, abstractmethod
from pathlib import Path
import traceback
from typing import Any, Dict, List, Union

from loguru import logger
from pydantic import BaseModel, ConfigDict
import torch

from fasr.config import Config
from fasr.data.audio import (
    Audio,
    AudioList,
)
from fasr.utils.base import IOMixin
from fasr.utils.load_utils import load_plugin


class BaseComponent(IOMixin, BaseModel, ABC):
    """A component is a module that can set tag on audio data"""

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)
    timer_data: Dict[str, float] | None = (
        None  # 使用utils.time_it.timer装饰器时，自动记录时间。
    )
    name: str | None = None
    input_tags: List[str] = []
    output_tags: List[str] = []

    @abstractmethod
    def predict(self, audios: AudioList) -> AudioList:
        raise NotImplementedError

    @abstractmethod
    def setup(
        self, checkpoint_dir: str, device: str = "cpu", compile: bool = False, **kwargs
    ):
        raise NotImplementedError

    def get_config(self) -> Config:
        raise NotImplementedError

    def load(self, save_dir: str):
        raise NotImplementedError

    def save(self, save_dir: str):
        raise NotImplementedError

    def log_component_error(self, details: str = ""):
        logger.error(
            f"Component {self.name} error, details: {details}",
        )

    def load_model_plugin(self, group: str, model: str):
        """通过 entry_points 加载指定的模型插件并注册到 registry。

        Args:
            group (str): entry_points 的组名，如 'fasr.asr_models'。
            model (str): 该组下的插件名称，如 'paraformer'。
        """
        load_plugin(group=group, name=model)

    def _filter_bad_audios(self, audios: AudioList) -> AudioList:
        """filter bad audios"""
        ids = []
        for audio in audios:
            audio: Audio
            if not audio.is_bad:
                ids.append(audio.id)
        return audios.filter_audio_id(ids)

    def _to_audios(
        self,
        input: Union[str, List[str], Any, Audio, AudioList, Path, List[Path]],
    ) -> AudioList:
        if isinstance(input, str):
            audios = AudioList.from_urls([input], load=False)
            return audios
        elif isinstance(input, list):
            audios = AudioList()
            for item in input:
                if isinstance(item, str):
                    audio = Audio(url=item)
                    audios.append(audio)
                elif isinstance(item, Audio):
                    audios.append(item)
                elif isinstance(item, Path):
                    audio = Audio(url=item)
                    audios.append(audio)
                else:
                    raise ValueError(
                        f"Invalid item type: {type(item)} for component {self.name}"
                    )
            return audios
        elif isinstance(input, Audio):
            return AudioList([input])
        elif isinstance(input, AudioList):
            return input
        elif isinstance(input, Path):
            return AudioList(docs=[Audio(url=input)])
        else:
            raise ValueError(
                f"Invalid input type: {type(input)} for component {self.name}"
            )

    def __ror__(
        self,
        input: Union[str, List[str], Audio, AudioList],
        *args,
        **kwargs,
    ) -> AudioList:
        """组件之间的同步连接符号 `|` 实现"""
        audios = self._to_audios(input=input)
        audios = self._filter_bad_audios(audios)
        for audio in audios:
            audio: Audio
            if audio.pipeline is None:
                audio.pipeline = []
            audio.pipeline.append(self.name)
        try:
            audios = self.predict(audios)
        except Exception as e:
            for audio in audios:
                audio.is_bad = True
                audio.bad_reason = str(e)
                audio.bad_component = self.name
                self.log_component_error(details=traceback.format_exc())
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
                current_gpu_index = torch.cuda.current_device()
                available_memory = torch.cuda.get_device_properties(
                    current_gpu_index
                ).total_memory / (1024**3)
                used_memory = torch.cuda.memory_allocated(current_gpu_index) / (1024**3)
                free_memory = available_memory - used_memory
                if free_memory <= 0:
                    raise MemoryError("Out of GPU memory.")
        return audios



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import re

import torch
from transformers import AutoTokenizer
from transformers.trainer_pt_utils import LabelSmoother

DEFAULT_SPEECH_TOKEN = "<speech>"
IGNORE_TOKEN_ID = LabelSmoother.ignore_index


class LlmTokenizerWrapper:
    @classmethod
    def build_llm_tokenizer(cls, llm_path, use_flash_attn=False):
        tokenizer = AutoTokenizer.from_pretrained(llm_path)
        if use_flash_attn:
            tokenizer.padding_side = "left"
        else:
            tokenizer.padding_side = "right"
        special_tokens_dict = {"additional_special_tokens": [DEFAULT_SPEECH_TOKEN]}
        tokenizer.add_special_tokens(special_tokens_dict)
        return tokenizer

    @classmethod
    def clean_text(cls, origin_text):
        """remove punc, remove space between Chinese and keep space between English"""
        # remove punc
        text = re.sub("[，。？！,\.!?《》（）\·“”、\\/]", "", origin_text)
        # merge space
        text = re.sub("\s+", " ", text)

        # remove space between Chinese and keep space between English
        pattern = re.compile(r"([\u3400-\u4dbf\u4e00-\u9fff])")  # Chinese
        parts = pattern.split(text.strip())
        parts = [p for p in parts if len(p.strip()) > 0]
        text = "".join(parts)
        text = text.strip()

        text = text.lower()
        return text

    @classmethod
    def preprocess_texts(cls, origin_texts, tokenizer, max_len, decode=False):
        messages = []
        clean_texts = []
        for i, origin_text in enumerate(origin_texts):
            text = cls.clean_text(origin_text)
            clean_texts.append(text)
            text = text if not decode else ""
            message = [
                {"role": "user", "content": f"{DEFAULT_SPEECH_TOKEN}请转写音频为文字"},
                {"role": "assistant", "content": text},
            ]
            messages.append(message)

        texts = []
        if not decode:
            TEMPLATE = "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content']}}{% if loop.last %}{{ '<|im_end|>'}}{% else %}{{ '<|im_end|>\n' }}{% endif %}{% endfor %}"
        else:
            TEMPLATE = "{% for message in messages %}{{'<|im_start|>' + message['role'] + '\n' + message['content']}}{% if loop.last %}{{''}}{% else %}{{ '<|im_end|>\n' }}{% endif %}{% endfor %}"
        for i, msg in enumerate(messages):
            texts.append(
                tokenizer.apply_chat_template(
                    msg,
                    tokenize=True,
                    chat_template=TEMPLATE,
                    add_generation_prompt=False,
                    max_length=max_len,
                    truncation=True,
                    return_dict=False,
                )
            )

        # Padding texts
        max_len_texts = max([len(text) for text in texts])
        if tokenizer.padding_side == "right":
            texts = [
                list(text) + [tokenizer.pad_token_id] * (max_len_texts - len(text))
                for text in texts
            ]
        else:
            texts = [
                [tokenizer.pad_token_id] * (max_len_texts - len(text)) + text
                for text in texts
            ]
        input_ids = torch.tensor(texts, dtype=torch.int)

        target_ids = input_ids.clone()
        target_ids[target_ids == tokenizer.pad_token_id] = IGNORE_TOKEN_ID

        # first get the indices of the tokens
        mask_prompt = True
        if mask_prompt:
            mask_indices = torch.where(
                input_ids == tokenizer.convert_tokens_to_ids("assistant")
            )
            for i in range(mask_indices[0].size(0)):
                row = mask_indices[0][i]
                col = mask_indices[1][i]
                target_ids[row, : col + 2] = IGNORE_TOKEN_ID

        attention_mask = input_ids.ne(tokenizer.pad_token_id)

        target_ids = target_ids.type(torch.LongTensor)
        input_ids = input_ids.type(torch.LongTensor)
        return input_ids, attention_mask, target_ids, clean_texts



# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu)

import logging
import re

import sentencepiece as spm

from ..data.token_dict import TokenDict


class ChineseCharEnglishSpmTokenizer:
    """
    - One Chinese char is a token.
    - Split English word into SPM and one piece is a token.
    - Ignore ' ' between Chinese char
    - Replace ' ' between English word with "▁" by spm_model
    - Need to put SPM piece into dict file
    - If not set spm_model, will use English char and <space>
    """

    SPM_SPACE = "▁"

    def __init__(self, dict_path, spm_model, unk="<unk>", space="<space>"):
        self.dict = TokenDict(dict_path, unk=unk)
        self.space = space
        if spm_model:
            self.sp = spm.SentencePieceProcessor()
            self.sp.Load(spm_model)
        else:
            self.sp = None
            print("[WRAN] Not set spm_model, will use English char")
            print("[WARN] Please check how to deal with ' '(space)")
            if self.space not in self.dict:
                print("Please add <space> to your dict, or it will be <unk>")

    def tokenize(self, text, replace_punc=True):
        # if text == "":
        #    logging.info(f"empty text")
        text = text.upper()
        tokens = []
        if replace_punc:
            text = re.sub("[，。？！,\.?!]", " ", text)
        pattern = re.compile(r"([\u3400-\u4dbf\u4e00-\u9fff])")
        parts = pattern.split(text.strip())
        parts = [p for p in parts if len(p.strip()) > 0]
        for part in parts:
            if pattern.fullmatch(part) is not None:
                tokens.append(part)
            else:
                if self.sp:
                    for piece in self.sp.EncodeAsPieces(part.strip()):
                        tokens.append(piece)
                else:
                    for char in part.strip():
                        tokens.append(char if char != " " else self.space)
        tokens_id = []
        for token in tokens:
            tokens_id.append(self.dict.get(token, self.dict.unk))
        return tokens, tokens_id

    def detokenize(self, inputs, join_symbol="", replace_spm_space=True):
        """inputs is ids or tokens, do not need self.sp"""
        if len(inputs) > 0 and type(inputs[0]) == int:
            tokens = [self.dict[id] for id in inputs]
        else:
            tokens = inputs
        s = f"{join_symbol}".join(tokens)
        if replace_spm_space:
            s = s.replace(self.SPM_SPACE, " ").strip()
        return s

    def merge_spm_timestamp(self, timestamp):
        merged_timestamp = []
        i = 0
        while i < len(timestamp):
            token, start, end = timestamp[i]
            if token.startswith(self.SPM_SPACE):
                token = token.replace(self.SPM_SPACE, "")
                current_end = end
                next_i = i + 1
                while next_i < len(timestamp):
                    next_token, next_start, next_end = timestamp[next_i]
                    if re.match("^[a-zA-Z']+$", next_token):
                        token += next_token
                        current_end = next_end
                        next_i += 1
                    else:
                        break
                end = current_end
                i = next_i
            else:
                i += 1
            merged_timestamp.append((token, start, end))
        return merged_timestamp



from pathlib import Path
import re
import string

from kaldifst import TextNormalizer as normalizer

DEFAULT_DIR = Path(__file__).parent / "asset" / "wetext"

EOS = "<EOS>"
TN_ORDERS = {
    "date": ["year", "month", "day"],
    "fraction": ["denominator", "numerator"],
    "measure": ["denominator", "numerator", "value"],
    "money": ["value", "currency"],
    "time": ["noon", "hour", "minute", "second"],
}
EN_TN_ORDERS = {
    "date": ["preserve_order", "text", "day", "month", "year"],
    "money": ["integer_part", "fractional_part", "quantity", "currency_maj"],
}
ITN_ORDERS = {
    "date": ["year", "month", "day"],
    "fraction": ["sign", "numerator", "denominator"],
    "measure": ["numerator", "denominator", "value"],
    "money": ["currency", "value", "decimal"],
    "time": ["hour", "minute", "second", "noon"],
}


class Token:
    def __init__(self, name):
        self.name = name
        self.order = []
        self.members = {}

    def append(self, key, value):
        self.order.append(key)
        self.members[key] = value

    def string(self, orders):
        output = self.name + " {"
        if self.name in orders.keys():
            if (
                "preserve_order" not in self.members.keys()
                or self.members["preserve_order"] != "true"
            ):
                self.order = orders[self.name]

        for key in self.order:
            if key not in self.members.keys():
                continue
            output += ' {}: "{}"'.format(key, self.members[key])
        return output + " }"


class TokenParser:
    def __init__(self, lang, operator="tn"):
        assert lang in ("en", "zh")
        if lang == "en":
            if operator == "tn":
                self.orders = EN_TN_ORDERS
            else:
                raise NotImplementedError()
        else:
            if operator == "tn":
                self.orders = TN_ORDERS
            elif operator == "itn":
                self.orders = ITN_ORDERS

    def load(self, input):
        assert len(input) > 0
        self.index = 0
        self.text = input
        self.char = input[0]
        self.tokens = []

    def read(self):
        if self.index < len(self.text) - 1:
            self.index += 1
            self.char = self.text[self.index]
            return True
        self.char = EOS
        return False

    def parse_ws(self):
        not_eos = self.char != EOS
        while not_eos and self.char == " ":
            not_eos = self.read()
        return not_eos

    def parse_char(self, exp):
        if self.char == exp:
            self.read()
            return True
        return False

    def parse_chars(self, exp):
        ok = False
        for x in exp:
            ok |= self.parse_char(x)
        return ok

    def parse_key(self):
        assert self.char != EOS
        assert self.char not in string.whitespace

        key = ""
        while self.char in string.ascii_letters + "_":
            key += self.char
            self.read()
        return key

    def parse_value(self):
        assert self.char != EOS
        escape = False

        value = ""
        while self.char != '"':
            value += self.char
            escape = self.char == "\\"
            self.read()
            if escape:
                escape = False
                value += self.char
                self.read()
        return value

    def parse(self, input):
        self.load(input)
        while self.parse_ws():
            name = self.parse_key()
            self.parse_chars(" { ")

            token = Token(name)
            while self.parse_ws():
                if self.char == "}":
                    self.parse_char("}")
                    break
                key = self.parse_key()
                self.parse_chars(': "')
                value = self.parse_value()
                self.parse_char('"')
                token.append(key, value)
            self.tokens.append(token)

    def reorder(self, input):
        self.parse(input)
        output = ""
        for token in self.tokens:
            output += token.string(self.orders) + " "
        return output.strip()


class Normalizer:
    def __init__(
        self,
        tagger_path=None,
        verbalizer_path=None,
        lang="auto",
        operator="tn",
        remove_erhua=False,
        enable_0_to_9=False,
    ):
        self.lang = lang
        self.operator = operator
        self.taggers = {}
        self.verbalizers = {}
        if tagger_path is None or verbalizer_path is None:
            repo_dir = DEFAULT_DIR
            assert lang in ("auto", "en", "zh") and operator in ("tn", "itn")

            taggers = {"en": "tagger.fst", "zh": "tagger.fst"}
            verbalizers = {"en": "verbalizer.fst", "zh": "verbalizer.fst"}
            if operator == "itn" and enable_0_to_9:
                taggers["zh"] = "tagger_enable_0_to_9.fst"
            if operator == "tn" and remove_erhua:
                verbalizers["zh"] = "verbalizer_remove_erhua.fst"

            for lang in ("en", "zh"):
                if self.lang in ("auto", lang):
                    self.taggers[lang] = normalizer(
                        str(repo_dir / lang / operator / taggers[lang])
                    )
                    self.verbalizers[lang] = normalizer(
                        str(repo_dir / lang / operator / verbalizers[lang])
                    )
        else:
            assert lang in (
                "en",
                "zh",
            ), "Language must be 'en' or 'zh' when using custom tagger and verbalizer."
            self.taggers[lang] = normalizer(tagger_path)
            self.verbalizers[lang] = normalizer(verbalizer_path)

    def tag(self, text, lang=None):
        lang = lang or self.lang
        return self.taggers[lang](text)

    def verbalize(self, text, lang=None):
        lang = lang or self.lang
        text = TokenParser(lang, self.operator).reorder(text)
        return self.verbalizers[lang](text)

    @staticmethod
    def contains_chinese(string):
        for ch in string:
            if "\u4e00" <= ch <= "\u9fff":
                return True
        return False

    def normalize(self, text):
        if self.operator == "itn" or bool(re.search(r"\d", text)):
            lang = self.lang
            if lang == "auto":
                lang = "zh" if Normalizer.contains_chinese(text) else "en"
            return self.verbalize(self.tag(text, lang), lang)
        return text



from pathlib import Path
import time
from typing import Dict, Iterable

from funasr import AutoModel
from funasr_onnx import Fsmn_vad_online
from loguru import logger
import numpy as np
from typing_extensions import Self

from fasr.config import registry
from fasr.data import AudioChunk, AudioSpan, Waveform
from fasr.model import StreamVADModel

DEFAULT_CHECKPOINT_DIR = Path(__file__).parent.parent / "asset" / "fsmn-vad"


@registry.stream_vad_models.register("stream_fsmn")
@registry.stream_vad_models.register("stream_fsmn.torch")
class FSMNForStreamVAD(StreamVADModel):
    fsmn: AutoModel | None = None
    chunk_size_ms: int = 100
    sample_rate: int = 16000
    max_end_silence_time: int = 500
    speech_noise_thres: float = 0.6
    db_threshold: int = -100

    def detect_chunk(
        self,
        chunk: AudioChunk,
    ) -> Iterable[AudioSpan]:
        """Detect voice activity in the given chunk of waveform.

        Args:
            waveform (Waveform): The chunk of waveform to detect voice activity in.
            is_last (bool): Indicates if the chunk is the last chunk of the audio.
            state (Dict): The state of the model, which includes buffer, cache, offset, is_detected, and audio_waveform.

        Notes:
            - The function processes the waveform in chunks of size `self.chunk_size`.
            - If the sample rate of the input waveform does not match `self.sample_rate`, it will be resampled.
            - The function maintains a buffer to handle waveform chunks and updates the state accordingly.
            - Voice activity detection is performed using the FSMN model, and detected segments are yielded as AudioSpan objects.
            - If `is_last` is True and there is remaining data in the buffer, it processes the final chunk.

        Yields:
            AudioSpan: Detected voice activity spans, each represented as an AudioSpan object.
        """
        if chunk.is_start and chunk.stream_id in self.states:
            logger.warning(
                "found start chunk with existing state, this is not expected"
            )
            del self.states[chunk.stream_id]
        state: Dict = self.states.get(chunk.stream_id, {})
        if chunk.waveform.sample_rate != self.sample_rate:
            chunk.waveform.resample(self.sample_rate)
        buffer: Waveform = state.get(
            "buffer",
            Waveform(data=np.array([], dtype=np.float32), sample_rate=self.sample_rate),
        )
        buffer.append(chunk.waveform)
        audio_waveform: Waveform = state.get(
            "audio_waveform",
            Waveform(data=np.array([], dtype=np.float32), sample_rate=self.sample_rate),
        )
        audio_waveform.append(chunk.waveform)
        cache = state.get("cache", {})
        offset = state.get("offset", 0)
        is_detected = state.get("is_detected", False)
        while len(buffer) >= self.chunk_size:
            chunk_waveform = buffer[: self.chunk_size]
            buffer = buffer[self.chunk_size :]
            data = chunk_waveform.data
            sample_rate = chunk_waveform.sample_rate
            segments = self.fsmn.generate(
                input=data,
                fs=sample_rate,
                chunk_size=self.chunk_size_ms,
                is_final=False,
                cache=cache,
            )[0]["value"]
            if len(segments) > 0:
                for segment in segments:
                    start, end = segment
                    if start != -1 and end == -1:
                        is_detected = True
                        start_idx = start * sample_rate // 1000
                        end_idx = len(data) + offset
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_start",
                            is_start=True,
                        )

                    if start == -1 and end != -1:
                        is_detected = False
                        start_idx = offset
                        end_idx = end * sample_rate // 1000
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_end",
                            is_last=True,
                        )

                    if start != -1 and end != -1:
                        is_detected = False
                        start_idx = start * sample_rate // 1000
                        end_idx = end * sample_rate // 1000
                        segment_waveform = audio_waveform[
                            start_idx:end_idx
                        ]  # start_idx 有可能大于 end_idx
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_mid",
                        )
            else:
                if is_detected:
                    yield AudioChunk(
                        stream_id=chunk.stream_id,
                        start_ms=-1,
                        end_ms=-1,
                        waveform=chunk_waveform,
                        sample_rate=sample_rate,
                        vad_state="segment_mid",
                    )

            offset += len(data)

        if chunk.is_last and len(buffer) > 0:
            chunk_waveform = buffer
            buffer = Waveform(data=np.array([]), sample_rate=self.sample_rate)
            data = chunk_waveform.data
            sample_rate = chunk_waveform.sample_rate
            segments = self.fsmn.generate(
                input=data,
                fs=sample_rate,
                is_final=True,
                cache=cache,
            )[0]["value"]
            if len(segments) > 0:
                for segment in segments:
                    start, end = segment
                    if start != -1 and end == -1:
                        is_detected = True
                        start_idx = start * sample_rate // 1000
                        end_idx = len(data) + offset
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_start",
                            is_start=True,
                        )

                    if start == -1 and end != -1:
                        is_detected = False
                        start_idx = offset
                        end_idx = end * sample_rate // 1000
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_end",
                            is_last=True,
                        )

                    if start != -1 and end != -1:
                        is_detected = False
                        start_idx = start * sample_rate // 1000
                        end_idx = end * sample_rate // 1000
                        segment_waveform = audio_waveform[
                            start_idx:end_idx
                        ]  # start_idx 有可能大于 end_idx
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_mid",
                        )
            else:
                if is_detected:
                    yield AudioChunk(
                        stream_id=chunk.stream_id,
                        start_ms=-1,
                        end_ms=-1,
                        waveform=chunk_waveform,
                        sample_rate=sample_rate,
                        vad_state="segment_mid",
                    )

            offset += len(data)

        if chunk.is_last and chunk.stream_id in self.states:
            del self.states[chunk.stream_id]
        else:
            state.update(
                {
                    "buffer": buffer,
                    "cache": cache,
                    "offset": offset,
                    "is_detected": is_detected,
                    "audio_waveform": audio_waveform,
                }
            )
            self.states[chunk.stream_id] = state

    def from_checkpoint(self, checkpoint_dir: str | None = None, **kwargs) -> Self:
        checkpoint_dir = checkpoint_dir or DEFAULT_CHECKPOINT_DIR
        self.fsmn = AutoModel(
            model=str(checkpoint_dir),
            disable_update=True,
            disable_log=True,
            disable_pbar=True,
            max_end_silence_time=self.max_end_silence_time,
            speech_noise_thres=self.speech_noise_thres,
            decibel_thres=self.db_threshold,
            **kwargs,
        )
        return self


@registry.stream_vad_models.register("stream_fsmn.onnx")
class FSMNForStreamVADOnnx(StreamVADModel):
    fsmn: Fsmn_vad_online | None = None
    chunk_size_ms: int = 100
    sample_rate: int = 16000
    max_end_silence_time: int = 500
    speech_noise_thres: float = 0.6

    def detect_chunk(self, chunk: AudioChunk) -> Iterable[AudioChunk]:
        """Detect voice activity in the given chunk of waveform.
        Args:
            waveform (Waveform): The chunk of waveform to detect voice activity in.
            is_last (bool): Indicates if the chunk is the last chunk of the audio.
            state (Dict): The state of the model, which includes buffer, cache, offset, is_detected, and audio_waveform.
        Notes:
            - The function processes the waveform in chunks of size `self.chunk_size`.
            - If the sample rate of the input waveform does not match `self.sample_rate`, it will be resampled.
            - The function maintains a buffer to handle waveform chunks and updates the state accordingly.
            - Voice activity detection is performed using the FSMN model, and detected segments are yielded as AudioSpan objects.
            - If `is_last` is True and there is remaining data in the buffer, it processes the final chunk.
        Yields:
            AudioSpan: Detected voice activity spans, each represented as an AudioSpan object.
        """
        if chunk.is_start and chunk.stream_id in self.states:
            logger.warning(
                "found start chunk with existing state, this is not expected"
            )
            del self.states[chunk.stream_id]
        state: Dict = self.states.get(chunk.stream_id, {})
        waveform = chunk.waveform
        if waveform.sample_rate != self.sample_rate:
            waveform = waveform.resample(self.sample_rate)
        buffer: Waveform = state.get(
            "buffer",
            Waveform(data=np.array([], dtype=np.float32), sample_rate=self.sample_rate),
        )
        buffer = buffer.append(waveform)
        audio_waveform: Waveform = state.get(
            "audio_waveform",
            Waveform(data=np.array([], dtype=np.float32), sample_rate=self.sample_rate),
        )
        audio_waveform = audio_waveform.append(waveform=waveform)
        in_cache = state.get("in_cache", [])
        offset = state.get("offset", 0)
        is_detected = state.get("is_detected", False)
        param_dict = {"in_cache": in_cache, "is_final": chunk.is_last}
        while len(buffer) >= self.chunk_size:
            chunk_waveform = buffer[: self.chunk_size]
            buffer = buffer[self.chunk_size :]
            data = chunk_waveform.data
            sample_rate = chunk_waveform.sample_rate
            start = time.perf_counter()
            segments = self.fsmn(audio_in=data, param_dict=param_dict)
            end = time.perf_counter()
            if len(segments) > 0:
                for segment in segments[
                    0
                ]:  # segments[0] is the result of the first input, batch size is 1
                    start, end = segment
                    if start != -1 and end == -1:
                        is_detected = True
                        start_idx = start * sample_rate // 1000
                        end_idx = len(data) + offset
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_start",
                            is_start=True,
                        )
                    if start == -1 and end != -1:
                        is_detected = False
                        start_idx = offset
                        end_idx = end * sample_rate // 1000
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_end",
                            is_last=True,
                        )
                    if start != -1 and end != -1:
                        is_detected = False
                        start_idx = start * sample_rate // 1000
                        end_idx = end * sample_rate // 1000
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_mid",
                        )
            else:
                if is_detected:
                    yield AudioChunk(
                        stream_id=chunk.stream_id,
                        start_ms=-1,
                        end_ms=-1,
                        waveform=waveform,
                        sample_rate=sample_rate,
                        vad_state="segment_mid",
                    )
            offset += len(data)

        if chunk.is_last and 0 < len(buffer):
            chunk_waveform = buffer
            buffer = Waveform(data=np.array([]), sample_rate=self.sample_rate)
            data = chunk_waveform.data
            sample_rate = chunk_waveform.sample_rate
            segments = self.fsmn(audio_in=data, param_dict=param_dict)
            if len(segments) > 0:
                for segment in segments[0]:
                    start, end = segment
                    if start != -1 and end == -1:
                        is_detected = True
                        start_idx = start * sample_rate // 1000
                        end_idx = len(data) + offset
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_start",
                            is_start=True,
                        )
                    if start == -1 and end != -1:
                        is_detected = False
                        start_idx = offset
                        end_idx = end * sample_rate // 1000
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_end",
                            is_last=True,
                        )
                    if start != -1 and end != -1:
                        is_detected = False
                        start_idx = start * sample_rate // 1000
                        end_idx = end * sample_rate // 1000
                        segment_waveform = audio_waveform[start_idx:end_idx]
                        yield AudioChunk(
                            stream_id=chunk.stream_id,
                            start_ms=start,
                            end_ms=end,
                            waveform=segment_waveform,
                            sample_rate=sample_rate,
                            vad_state="segment_mid",
                        )
            else:
                if is_detected:
                    yield AudioChunk(
                        stream_id=chunk.stream_id,
                        start_ms=-1,
                        end_ms=-1,
                        waveform=waveform,
                        sample_rate=sample_rate,
                        vad_state="segment_mid",
                    )
            offset += len(data)
        if chunk.is_last and chunk.stream_id in self.states:
            del self.states[chunk.stream_id]
        else:
            state.update(
                {
                    "buffer": buffer,
                    "in_cache": in_cache,
                    "offset": offset,
                    "is_detected": is_detected,
                    "audio_waveform": audio_waveform,
                }
            )
            self.states[chunk.stream_id] = state

    def from_checkpoint(
        self, checkpoint_dir: str | None = None, device: int | None = None, **kwargs
    ) -> Self:
        if not device:
            import torch

            device = "cuda" if torch.cuda.is_available() else "cpu"
            if device == "cuda":
                device_id = 0
            else:
                device_id = -1
        checkpoint_dir = checkpoint_dir or DEFAULT_CHECKPOINT_DIR
        self.fsmn = Fsmn_vad_online(
            model_dir=str(checkpoint_dir),
            max_end_sil=self.max_end_silence_time,
            speech_noise_thres=self.speech_noise_thres,
            device_id=device_id,
            **kwargs,
        )
        return self
